import inspect
from unittest import mock

import pytest

import sentry_sdk
from sentry_sdk.consts import SPANTEMPLATE
from sentry_sdk.tracing_utils import create_span_decorator
from sentry_sdk.utils import logger
from tests.conftest import patch_start_tracing_child


def my_example_function():
    return "return_of_sync_function"


async def my_async_example_function():
    return "return_of_async_function"


@pytest.mark.forked
def test_trace_decorator():
    with patch_start_tracing_child() as fake_start_child:
        result = my_example_function()
        fake_start_child.assert_not_called()
        assert result == "return_of_sync_function"

        start_child_span_decorator = create_span_decorator()
        result2 = start_child_span_decorator(my_example_function)()
        fake_start_child.assert_called_once_with(
            op="function", name="test_decorator.my_example_function"
        )
        assert result2 == "return_of_sync_function"


def test_trace_decorator_no_trx():
    with patch_start_tracing_child(fake_transaction_is_none=True):
        with mock.patch.object(logger, "debug", mock.Mock()) as fake_debug:
            result = my_example_function()
            fake_debug.assert_not_called()
            assert result == "return_of_sync_function"

            start_child_span_decorator = create_span_decorator()
            result2 = start_child_span_decorator(my_example_function)()
            fake_debug.assert_called_once_with(
                "Cannot create a child span for %s. "
                "Please start a Sentry transaction before calling this function.",
                "test_decorator.my_example_function",
            )
            assert result2 == "return_of_sync_function"


@pytest.mark.forked
@pytest.mark.asyncio
async def test_trace_decorator_async():
    with patch_start_tracing_child() as fake_start_child:
        result = await my_async_example_function()
        fake_start_child.assert_not_called()
        assert result == "return_of_async_function"

        start_child_span_decorator = create_span_decorator()
        result2 = await start_child_span_decorator(my_async_example_function)()
        fake_start_child.assert_called_once_with(
            op="function",
            name="test_decorator.my_async_example_function",
        )
        assert result2 == "return_of_async_function"


@pytest.mark.asyncio
async def test_trace_decorator_async_no_trx():
    with patch_start_tracing_child(fake_transaction_is_none=True):
        with mock.patch.object(logger, "debug", mock.Mock()) as fake_debug:
            result = await my_async_example_function()
            fake_debug.assert_not_called()
            assert result == "return_of_async_function"

            start_child_span_decorator = create_span_decorator()
            result2 = await start_child_span_decorator(my_async_example_function)()
            fake_debug.assert_any_call(
                "Cannot create a child span for %s. "
                "Please start a Sentry transaction before calling this function.",
                "test_decorator.my_async_example_function",
            )
            assert result2 == "return_of_async_function"


def test_trace_decorator_span_streaming(sentry_init, capture_items):
    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    items = capture_items("span")

    @sentry_sdk.traces.trace
    def traced_function():
        return "ok"

    result = traced_function()
    assert result == "ok"

    sentry_sdk.get_client().flush()
    spans = [item.payload for item in items]

    assert len(spans) == 1
    (span,) = spans

    assert (
        span["name"]
        == "test_decorator.test_trace_decorator_span_streaming.<locals>.traced_function"
    )
    assert span["attributes"]["sentry.op"] == "function"
    assert span["status"] == "ok"


def test_trace_decorator_arguments_span_streaming(sentry_init, capture_items):
    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    items = capture_items("span")

    @sentry_sdk.traces.trace(name="traced", attributes={"traced.attribute": 123})
    def traced_function():
        return "ok"

    result = traced_function()
    assert result == "ok"

    sentry_sdk.get_client().flush()
    spans = [item.payload for item in items]

    assert len(spans) == 1
    (span,) = spans

    assert span["name"] == "traced"
    assert span["attributes"]["traced.attribute"] == 123
    assert span["attributes"]["sentry.op"] == "function"
    assert span["status"] == "ok"


def test_trace_decorator_inactive_span_streaming(sentry_init, capture_items):
    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    items = capture_items("span")

    @sentry_sdk.traces.trace(name="outer", active=False)
    def traced_function():
        with sentry_sdk.traces.start_span(name="inner"):
            return "ok"

    result = traced_function()
    assert result == "ok"

    sentry_sdk.get_client().flush()
    spans = [item.payload for item in items]

    assert len(spans) == 2
    (span1, span2) = spans

    assert span1["name"] == "inner"
    assert span1.get("parent_span_id") != span2["span_id"]

    assert span2["name"] == "outer"


@pytest.mark.asyncio
async def test_trace_decorator_async_span_streaming(sentry_init, capture_items):
    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    items = capture_items("span")

    @sentry_sdk.traces.trace
    async def traced_function():
        return "ok"

    result = await traced_function()
    assert result == "ok"

    sentry_sdk.get_client().flush()
    spans = [item.payload for item in items]

    assert len(spans) == 1
    (span,) = spans

    assert (
        span["name"]
        == "test_decorator.test_trace_decorator_async_span_streaming.<locals>.traced_function"
    )
    assert span["attributes"]["sentry.op"] == "function"
    assert span["status"] == "ok"


@pytest.mark.asyncio
async def test_trace_decorator_async_arguments_span_streaming(
    sentry_init, capture_items
):
    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    items = capture_items("span")

    @sentry_sdk.traces.trace(name="traced", attributes={"traced.attribute": 123})
    async def traced_function():
        return "ok"

    result = await traced_function()
    assert result == "ok"

    sentry_sdk.get_client().flush()
    spans = [item.payload for item in items]

    assert len(spans) == 1
    (span,) = spans

    assert span["name"] == "traced"
    assert span["attributes"]["traced.attribute"] == 123
    assert span["attributes"]["sentry.op"] == "function"
    assert span["status"] == "ok"


@pytest.mark.asyncio
async def test_trace_decorator_async_inactive_span_streaming(
    sentry_init, capture_items
):
    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    items = capture_items("span")

    @sentry_sdk.traces.trace(name="outer", active=False)
    async def traced_function():
        with sentry_sdk.traces.start_span(name="inner"):
            return "ok"

    result = await traced_function()
    assert result == "ok"

    sentry_sdk.get_client().flush()
    spans = [item.payload for item in items]

    assert len(spans) == 2
    (span1, span2) = spans

    assert span1["name"] == "inner"
    assert span1.get("parent_span_id") != span2["span_id"]

    assert span2["name"] == "outer"


def test_trace_decorator_child_span_streaming(sentry_init, capture_items):
    """Spans created with @trace show up as children if a span is active."""
    sentry_init(
        traces_sample_rate=1.0,
        _experiments={
            "trace_lifecycle": "stream",
        },
    )

    items = capture_items("span")

    @sentry_sdk.traces.trace
    def _some_function_traced_stream(a, b, c):
        return True

    with sentry_sdk.traces.start_span(name="segment") as segment:
        result = _some_function_traced_stream(1, 2, 3)

    assert result is True

    sentry_sdk.flush()

    assert len(items) == 2
    child_span, segment_span = items[0].payload, items[1].payload

    assert (
        child_span["name"]
        == "test_decorator.test_trace_decorator_child_span_streaming.<locals>._some_function_traced_stream"
    )
    assert child_span["parent_span_id"] == segment.span_id
    assert segment_span["name"] == "segment"
    assert "parent_span_id" not in segment_span


@pytest.mark.asyncio
async def test_trace_decorator_async_child_span_streaming(sentry_init, capture_items):
    """Spans created with @trace show up as children if a span is active."""
    sentry_init(
        traces_sample_rate=1.0,
        _experiments={
            "trace_lifecycle": "stream",
        },
    )

    items = capture_items("span")

    @sentry_sdk.traces.trace
    async def _some_function_traced_stream(a, b, c):
        return True

    with sentry_sdk.traces.start_span(name="segment") as segment:
        result = await _some_function_traced_stream(1, 2, 3)

    assert result is True

    sentry_sdk.flush()

    assert len(items) == 2
    child_span, segment_span = items[0].payload, items[1].payload

    assert (
        child_span["name"]
        == "test_decorator.test_trace_decorator_async_child_span_streaming.<locals>._some_function_traced_stream"
    )
    assert child_span["parent_span_id"] == segment.span_id
    assert segment_span["name"] == "segment"
    assert "parent_span_id" not in segment_span


def test_functions_to_trace_signature_unchanged_sync(sentry_init):
    sentry_init(
        traces_sample_rate=1.0,
    )

    def _some_function(a, b, c):
        pass

    @sentry_sdk.trace
    def _some_function_traced(a, b, c):
        pass

    @sentry_sdk.traces.trace
    def _some_function_traced_stream(a, b, c):
        pass

    assert inspect.getcallargs(_some_function, 1, 2, 3) == inspect.getcallargs(
        _some_function_traced, 1, 2, 3
    )

    assert inspect.getcallargs(_some_function, 1, 2, 3) == inspect.getcallargs(
        _some_function_traced_stream, 1, 2, 3
    )


@pytest.mark.asyncio
async def test_functions_to_trace_signature_unchanged_async(sentry_init):
    sentry_init(
        traces_sample_rate=1.0,
    )

    async def _some_function(a, b, c):
        pass

    @sentry_sdk.trace
    async def _some_function_traced(a, b, c):
        pass

    @sentry_sdk.traces.trace
    async def _some_function_traced_stream(a, b, c):
        pass

    assert inspect.getcallargs(_some_function, 1, 2, 3) == inspect.getcallargs(
        _some_function_traced, 1, 2, 3
    )
    assert inspect.getcallargs(_some_function, 1, 2, 3) == inspect.getcallargs(
        _some_function_traced_stream, 1, 2, 3
    )


@pytest.mark.parametrize("stream_gen_ai_spans", [True, False])
def test_span_templates_ai_dicts(
    sentry_init,
    capture_events,
    capture_items,
    stream_gen_ai_spans,
):
    @sentry_sdk.trace(template=SPANTEMPLATE.AI_TOOL)
    def my_tool(arg1, arg2):
        return {
            "output": "my_tool_result",
            "usage": {
                "prompt_tokens": 10,
                "completion_tokens": 20,
                "total_tokens": 30,
            },
        }

    @sentry_sdk.trace(template=SPANTEMPLATE.AI_CHAT)
    def my_chat(model=None, **kwargs):
        return {
            "content": "my_chat_result",
            "usage": {
                "input_tokens": 11,
                "output_tokens": 22,
                "total_tokens": 33,
            },
            "model": f"{model}-v123",
        }

    @sentry_sdk.trace(template=SPANTEMPLATE.AI_AGENT)
    def my_agent():
        my_tool(1, 2)
        my_chat(
            model="my-gpt-4o-mini",
            prompt="What is the weather in Tokyo?",
            system_prompt="You are a helpful assistant that can answer questions about the weather.",
            max_tokens=100,
            temperature=0.5,
            top_p=0.9,
            top_k=40,
            frequency_penalty=1.0,
            presence_penalty=2.0,
        )

    if stream_gen_ai_spans:
        sentry_init(
            traces_sample_rate=1.0,
            stream_gen_ai_spans=stream_gen_ai_spans,
        )
        items = capture_items("span")

        with sentry_sdk.start_transaction(name="test-transaction"):
            my_agent()

        (agent_span, tool_span, chat_span) = (
            item.payload for item in items if item.type == "span"
        )

        assert (
            agent_span["name"]
            == "invoke_agent test_decorator.test_span_templates_ai_dicts.<locals>.my_agent"
        )
        assert agent_span["attributes"] == {
            "gen_ai.agent.name": "test_decorator.test_span_templates_ai_dicts.<locals>.my_agent",
            "gen_ai.operation.name": "invoke_agent",
            "process.runtime.name": mock.ANY,
            "process.runtime.version": mock.ANY,
            "sentry.environment": "production",
            "sentry.op": "gen_ai.invoke_agent",
            "sentry.origin": "manual",
            "sentry.release": mock.ANY,
            "sentry.sdk.name": "sentry.python",
            "sentry.sdk.version": mock.ANY,
            "sentry.segment.id": mock.ANY,
            "sentry.segment.name": "test-transaction",
            "server.address": mock.ANY,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }

        assert (
            tool_span["name"]
            == "execute_tool test_decorator.test_span_templates_ai_dicts.<locals>.my_tool"
        )
        assert tool_span["attributes"] == {
            "gen_ai.tool.name": "test_decorator.test_span_templates_ai_dicts.<locals>.my_tool",
            "gen_ai.operation.name": "execute_tool",
            "gen_ai.usage.input_tokens": 10,
            "gen_ai.usage.output_tokens": 20,
            "gen_ai.usage.total_tokens": 30,
            "process.runtime.name": mock.ANY,
            "process.runtime.version": mock.ANY,
            "sentry.environment": "production",
            "sentry.op": "gen_ai.execute_tool",
            "sentry.origin": "manual",
            "sentry.release": mock.ANY,
            "sentry.sdk.name": "sentry.python",
            "sentry.sdk.version": mock.ANY,
            "sentry.segment.id": mock.ANY,
            "sentry.segment.name": "test-transaction",
            "server.address": mock.ANY,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }
        assert "gen_ai.tool.description" not in tool_span["attributes"]

        assert chat_span["name"] == "chat my-gpt-4o-mini"
        assert chat_span["attributes"] == {
            "gen_ai.operation.name": "chat",
            "gen_ai.request.frequency_penalty": 1.0,
            "gen_ai.request.max_tokens": 100,
            "gen_ai.request.messages": "[{'role': 'user', 'content': 'What is the weather in Tokyo?'}, {'role': 'system', 'content': 'You are a helpful assistant that can answer questions about the weather.'}]",
            "gen_ai.request.model": "my-gpt-4o-mini",
            "gen_ai.request.presence_penalty": 2.0,
            "gen_ai.request.temperature": 0.5,
            "gen_ai.request.top_k": 40,
            "gen_ai.request.top_p": 0.9,
            "gen_ai.response.model": "my-gpt-4o-mini-v123",
            "gen_ai.usage.input_tokens": 11,
            "gen_ai.usage.output_tokens": 22,
            "gen_ai.usage.total_tokens": 33,
            "process.runtime.name": mock.ANY,
            "process.runtime.version": mock.ANY,
            "sentry.environment": "production",
            "sentry.op": "gen_ai.chat",
            "sentry.origin": "manual",
            "sentry.release": mock.ANY,
            "sentry.sdk.name": "sentry.python",
            "sentry.sdk.version": mock.ANY,
            "sentry.segment.id": mock.ANY,
            "sentry.segment.name": "test-transaction",
            "server.address": mock.ANY,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }
    else:
        sentry_init(
            traces_sample_rate=1.0,
            stream_gen_ai_spans=stream_gen_ai_spans,
        )
        events = capture_events()

        with sentry_sdk.start_transaction(name="test-transaction"):
            my_agent()

        (event,) = events
        (agent_span, tool_span, chat_span) = event["spans"]

        assert agent_span["op"] == "gen_ai.invoke_agent"
        assert (
            agent_span["description"]
            == "invoke_agent test_decorator.test_span_templates_ai_dicts.<locals>.my_agent"
        )
        assert agent_span["data"] == {
            "gen_ai.agent.name": "test_decorator.test_span_templates_ai_dicts.<locals>.my_agent",
            "gen_ai.operation.name": "invoke_agent",
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }

        assert tool_span["op"] == "gen_ai.execute_tool"
        assert (
            tool_span["description"]
            == "execute_tool test_decorator.test_span_templates_ai_dicts.<locals>.my_tool"
        )
        assert tool_span["data"] == {
            "gen_ai.tool.name": "test_decorator.test_span_templates_ai_dicts.<locals>.my_tool",
            "gen_ai.operation.name": "execute_tool",
            "gen_ai.usage.input_tokens": 10,
            "gen_ai.usage.output_tokens": 20,
            "gen_ai.usage.total_tokens": 30,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }
        assert "gen_ai.tool.description" not in tool_span["data"]

        assert chat_span["op"] == "gen_ai.chat"
        assert chat_span["description"] == "chat my-gpt-4o-mini"
        assert chat_span["data"] == {
            "gen_ai.operation.name": "chat",
            "gen_ai.request.frequency_penalty": 1.0,
            "gen_ai.request.max_tokens": 100,
            "gen_ai.request.messages": "[{'role': 'user', 'content': 'What is the weather in Tokyo?'}, {'role': 'system', 'content': 'You are a helpful assistant that can answer questions about the weather.'}]",
            "gen_ai.request.model": "my-gpt-4o-mini",
            "gen_ai.request.presence_penalty": 2.0,
            "gen_ai.request.temperature": 0.5,
            "gen_ai.request.top_k": 40,
            "gen_ai.request.top_p": 0.9,
            "gen_ai.response.model": "my-gpt-4o-mini-v123",
            "gen_ai.usage.input_tokens": 11,
            "gen_ai.usage.output_tokens": 22,
            "gen_ai.usage.total_tokens": 33,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }


@pytest.mark.parametrize("stream_gen_ai_spans", [True, False])
def test_span_templates_ai_objects(
    sentry_init,
    capture_events,
    capture_items,
    stream_gen_ai_spans,
):
    @sentry_sdk.trace(template=SPANTEMPLATE.AI_TOOL)
    def my_tool(arg1, arg2):
        """This is a tool function."""
        mock_usage = mock.Mock()
        mock_usage.prompt_tokens = 10
        mock_usage.completion_tokens = 20
        mock_usage.total_tokens = 30

        mock_result = mock.Mock()
        mock_result.output = "my_tool_result"
        mock_result.usage = mock_usage

        return mock_result

    @sentry_sdk.trace(template=SPANTEMPLATE.AI_CHAT)
    def my_chat(model=None, **kwargs):
        mock_result = mock.Mock()
        mock_result.content = "my_chat_result"
        mock_result.usage = mock.Mock(
            input_tokens=11,
            output_tokens=22,
            total_tokens=33,
        )
        mock_result.model = f"{model}-v123"

        return mock_result

    @sentry_sdk.trace(template=SPANTEMPLATE.AI_AGENT)
    def my_agent():
        my_tool(1, 2)
        my_chat(
            model="my-gpt-4o-mini",
            prompt="What is the weather in Tokyo?",
            system_prompt="You are a helpful assistant that can answer questions about the weather.",
            max_tokens=100,
            temperature=0.5,
            top_p=0.9,
            top_k=40,
            frequency_penalty=1.0,
            presence_penalty=2.0,
        )

    sentry_init(
        traces_sample_rate=1.0,
        stream_gen_ai_spans=stream_gen_ai_spans,
    )

    if stream_gen_ai_spans:
        items = capture_items("span")

        with sentry_sdk.start_transaction(name="test-transaction"):
            my_agent()

        (agent_span, tool_span, chat_span) = (
            item.payload for item in items if item.type == "span"
        )

        assert (
            agent_span["name"]
            == "invoke_agent test_decorator.test_span_templates_ai_objects.<locals>.my_agent"
        )
        assert agent_span["attributes"] == {
            "gen_ai.agent.name": "test_decorator.test_span_templates_ai_objects.<locals>.my_agent",
            "gen_ai.operation.name": "invoke_agent",
            "process.runtime.name": mock.ANY,
            "process.runtime.version": mock.ANY,
            "sentry.environment": "production",
            "sentry.op": "gen_ai.invoke_agent",
            "sentry.origin": "manual",
            "sentry.release": mock.ANY,
            "sentry.sdk.name": "sentry.python",
            "sentry.sdk.version": mock.ANY,
            "sentry.segment.id": mock.ANY,
            "sentry.segment.name": "test-transaction",
            "server.address": mock.ANY,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }

        assert (
            tool_span["name"]
            == "execute_tool test_decorator.test_span_templates_ai_objects.<locals>.my_tool"
        )
        assert tool_span["attributes"] == {
            "gen_ai.tool.name": "test_decorator.test_span_templates_ai_objects.<locals>.my_tool",
            "gen_ai.tool.description": "This is a tool function.",
            "gen_ai.operation.name": "execute_tool",
            "gen_ai.usage.input_tokens": 10,
            "gen_ai.usage.output_tokens": 20,
            "gen_ai.usage.total_tokens": 30,
            "process.runtime.name": mock.ANY,
            "process.runtime.version": mock.ANY,
            "sentry.environment": "production",
            "sentry.op": "gen_ai.execute_tool",
            "sentry.origin": "manual",
            "sentry.release": mock.ANY,
            "sentry.sdk.name": "sentry.python",
            "sentry.sdk.version": mock.ANY,
            "sentry.segment.id": mock.ANY,
            "sentry.segment.name": "test-transaction",
            "server.address": mock.ANY,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }

        assert chat_span["name"] == "chat my-gpt-4o-mini"
        assert chat_span["attributes"] == {
            "gen_ai.operation.name": "chat",
            "gen_ai.request.frequency_penalty": 1.0,
            "gen_ai.request.max_tokens": 100,
            "gen_ai.request.messages": "[{'role': 'user', 'content': 'What is the weather in Tokyo?'}, {'role': 'system', 'content': 'You are a helpful assistant that can answer questions about the weather.'}]",
            "gen_ai.request.model": "my-gpt-4o-mini",
            "gen_ai.request.presence_penalty": 2.0,
            "gen_ai.request.temperature": 0.5,
            "gen_ai.request.top_k": 40,
            "gen_ai.request.top_p": 0.9,
            "gen_ai.response.model": "my-gpt-4o-mini-v123",
            "gen_ai.usage.input_tokens": 11,
            "gen_ai.usage.output_tokens": 22,
            "gen_ai.usage.total_tokens": 33,
            "process.runtime.name": mock.ANY,
            "process.runtime.version": mock.ANY,
            "sentry.environment": "production",
            "sentry.op": "gen_ai.chat",
            "sentry.origin": "manual",
            "sentry.release": mock.ANY,
            "sentry.sdk.name": "sentry.python",
            "sentry.sdk.version": mock.ANY,
            "sentry.segment.id": mock.ANY,
            "sentry.segment.name": "test-transaction",
            "server.address": mock.ANY,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }
    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test-transaction"):
            my_agent()

        (event,) = events
        (agent_span, tool_span, chat_span) = event["spans"]

        assert agent_span["op"] == "gen_ai.invoke_agent"
        assert (
            agent_span["description"]
            == "invoke_agent test_decorator.test_span_templates_ai_objects.<locals>.my_agent"
        )
        assert agent_span["data"] == {
            "gen_ai.agent.name": "test_decorator.test_span_templates_ai_objects.<locals>.my_agent",
            "gen_ai.operation.name": "invoke_agent",
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }

        assert tool_span["op"] == "gen_ai.execute_tool"
        assert (
            tool_span["description"]
            == "execute_tool test_decorator.test_span_templates_ai_objects.<locals>.my_tool"
        )
        assert tool_span["data"] == {
            "gen_ai.tool.name": "test_decorator.test_span_templates_ai_objects.<locals>.my_tool",
            "gen_ai.tool.description": "This is a tool function.",
            "gen_ai.operation.name": "execute_tool",
            "gen_ai.usage.input_tokens": 10,
            "gen_ai.usage.output_tokens": 20,
            "gen_ai.usage.total_tokens": 30,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }

        assert chat_span["op"] == "gen_ai.chat"
        assert chat_span["description"] == "chat my-gpt-4o-mini"
        assert chat_span["data"] == {
            "gen_ai.operation.name": "chat",
            "gen_ai.request.frequency_penalty": 1.0,
            "gen_ai.request.max_tokens": 100,
            "gen_ai.request.messages": "[{'role': 'user', 'content': 'What is the weather in Tokyo?'}, {'role': 'system', 'content': 'You are a helpful assistant that can answer questions about the weather.'}]",
            "gen_ai.request.model": "my-gpt-4o-mini",
            "gen_ai.request.presence_penalty": 2.0,
            "gen_ai.request.temperature": 0.5,
            "gen_ai.request.top_k": 40,
            "gen_ai.request.top_p": 0.9,
            "gen_ai.response.model": "my-gpt-4o-mini-v123",
            "gen_ai.usage.input_tokens": 11,
            "gen_ai.usage.output_tokens": 22,
            "gen_ai.usage.total_tokens": 33,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }


@pytest.mark.parametrize("stream_gen_ai_spans", [True, False])
@pytest.mark.parametrize("send_default_pii", [True, False])
def test_span_templates_ai_pii(
    sentry_init,
    capture_events,
    capture_items,
    send_default_pii,
    stream_gen_ai_spans,
):
    @sentry_sdk.trace(template=SPANTEMPLATE.AI_TOOL)
    def my_tool(arg1, arg2, **kwargs):
        """This is a tool function."""
        return "tool_output"

    @sentry_sdk.trace(template=SPANTEMPLATE.AI_CHAT)
    def my_chat(model=None, **kwargs):
        return "chat_output"

    @sentry_sdk.trace(template=SPANTEMPLATE.AI_AGENT)
    def my_agent(*args, **kwargs):
        my_tool(1, 2, tool_arg1="3", tool_arg2="4")
        my_chat(
            model="my-gpt-4o-mini",
            prompt="What is the weather in Tokyo?",
            system_prompt="You are a helpful assistant that can answer questions about the weather.",
            max_tokens=100,
            temperature=0.5,
            top_p=0.9,
            top_k=40,
            frequency_penalty=1.0,
            presence_penalty=2.0,
        )
        return "agent_output"

    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        stream_gen_ai_spans=stream_gen_ai_spans,
    )

    if stream_gen_ai_spans:
        items = capture_items("span")

        with sentry_sdk.start_transaction(name="test-transaction"):
            my_agent(22, 33, arg1=44, arg2=55)

        (_, tool_span, _) = (item.payload for item in items if item.type == "span")

        if send_default_pii:
            assert (
                tool_span["attributes"]["gen_ai.tool.input"]
                == "{'args': (1, 2), 'kwargs': {'tool_arg1': '3', 'tool_arg2': '4'}}"
            )
            assert tool_span["attributes"]["gen_ai.tool.output"] == "'tool_output'"
        else:
            assert "gen_ai.tool.input" not in tool_span["attributes"]
            assert "gen_ai.tool.output" not in tool_span["attributes"]
    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test-transaction"):
            my_agent(22, 33, arg1=44, arg2=55)

        (event,) = events
        (_, tool_span, _) = event["spans"]

        if send_default_pii:
            assert (
                tool_span["data"]["gen_ai.tool.input"]
                == "{'args': (1, 2), 'kwargs': {'tool_arg1': '3', 'tool_arg2': '4'}}"
            )
            assert tool_span["data"]["gen_ai.tool.output"] == "'tool_output'"
        else:
            assert "gen_ai.tool.input" not in tool_span["data"]
            assert "gen_ai.tool.output" not in tool_span["data"]



from unittest.mock import MagicMock, patch

import pytest

from sentry_sdk.integrations.opentelemetry.integration import OpenTelemetryIntegration


@pytest.mark.forked
def test_integration_enabled_if_option_is_on(sentry_init, reset_integrations):
    mocked_setup_sentry_tracing = MagicMock()

    with patch(
        "sentry_sdk.integrations.opentelemetry.integration._setup_sentry_tracing",
        mocked_setup_sentry_tracing,
    ):
        with pytest.warns(DeprecationWarning):
            sentry_init(_experiments={"otel_powered_performance": True})
        mocked_setup_sentry_tracing.assert_called_once()


@pytest.mark.forked
def test_integration_not_enabled_if_option_is_off(sentry_init, reset_integrations):
    mocked_setup_sentry_tracing = MagicMock()

    with patch(
        "sentry_sdk.integrations.opentelemetry.integration._setup_sentry_tracing",
        mocked_setup_sentry_tracing,
    ):
        sentry_init(_experiments={"otel_powered_performance": False})
        mocked_setup_sentry_tracing.assert_not_called()


@pytest.mark.forked
def test_integration_not_enabled_if_option_is_missing(sentry_init, reset_integrations):
    mocked_setup_sentry_tracing = MagicMock()

    with patch(
        "sentry_sdk.integrations.opentelemetry.integration._setup_sentry_tracing",
        mocked_setup_sentry_tracing,
    ):
        sentry_init()
        mocked_setup_sentry_tracing.assert_not_called()


@pytest.mark.forked
def test_integration_disabled_with_span_streaming(sentry_init, reset_integrations):
    mocked_setup_sentry_tracing = MagicMock()

    with patch(
        "sentry_sdk.integrations.opentelemetry.integration._setup_sentry_tracing",
        mocked_setup_sentry_tracing,
    ):
        sentry_init(
            integrations=[OpenTelemetryIntegration()],
            trace_lifecycle="stream",
        )
        mocked_setup_sentry_tracing.assert_not_called()



import sentry_sdk
from sentry_sdk.tracing import NoOpSpan

# These tests make sure that the examples from the documentation [1]
# are working when OTel (OpenTelemetry) instrumentation is turned on,
# and therefore, the Sentry tracing should not do anything.
#
# 1: https://docs.sentry.io/platforms/python/performance/instrumentation/custom-instrumentation/


def test_noop_start_transaction(sentry_init):
    sentry_init(instrumenter="otel")

    with sentry_sdk.start_transaction(
        op="task", name="test_transaction_name"
    ) as transaction:
        assert isinstance(transaction, NoOpSpan)
        assert sentry_sdk.get_current_scope().span is transaction

        transaction.name = "new name"


def test_noop_start_span(sentry_init):
    sentry_init(instrumenter="otel")

    with sentry_sdk.start_span(op="http", name="GET /") as span:
        assert isinstance(span, NoOpSpan)
        assert sentry_sdk.get_current_scope().span is span

        span.set_tag("http.response.status_code", 418)
        span.set_data("http.entity_type", "teapot")


def test_noop_transaction_start_child(sentry_init):
    sentry_init(instrumenter="otel")

    transaction = sentry_sdk.start_transaction(name="task")
    assert isinstance(transaction, NoOpSpan)

    with transaction.start_child(op="child_task") as child:
        assert isinstance(child, NoOpSpan)
        assert sentry_sdk.get_current_scope().span is child


def test_noop_span_start_child(sentry_init):
    sentry_init(instrumenter="otel")
    span = sentry_sdk.start_span(name="task")
    assert isinstance(span, NoOpSpan)

    with span.start_child(op="child_task") as child:
        assert isinstance(child, NoOpSpan)
        assert sentry_sdk.get_current_scope().span is child



import gc
import sys
from concurrent import futures
from textwrap import dedent
from threading import Thread

import pytest

import sentry_sdk
from sentry_sdk import capture_message
from sentry_sdk.integrations.threading import ThreadingIntegration

original_start = Thread.start
original_run = Thread.run


@pytest.mark.filterwarnings("ignore:.*:pytest.PytestUnhandledThreadExceptionWarning")
@pytest.mark.parametrize("integrations", [[ThreadingIntegration()], []])
def test_handles_exceptions(sentry_init, capture_events, integrations):
    sentry_init(default_integrations=False, integrations=integrations)
    events = capture_events()

    def crash():
        1 / 0

    t = Thread(target=crash)
    t.start()
    t.join()

    if integrations:
        (event,) = events

        (exception,) = event["exception"]["values"]
        assert exception["type"] == "ZeroDivisionError"
        assert exception["mechanism"]["type"] == "threading"
        assert not exception["mechanism"]["handled"]
    else:
        assert not events


@pytest.mark.filterwarnings("ignore:.*:pytest.PytestUnhandledThreadExceptionWarning")
@pytest.mark.parametrize("propagate_hub", (True, False))
def test_propagates_hub(sentry_init, capture_events, propagate_hub):
    sentry_init(
        default_integrations=False,
        integrations=[ThreadingIntegration(propagate_hub=propagate_hub)],
    )
    events = capture_events()

    def stage1():
        sentry_sdk.get_isolation_scope().set_tag("stage1", "true")

        t = Thread(target=stage2)
        t.start()
        t.join()

    def stage2():
        1 / 0

    t = Thread(target=stage1)
    t.start()
    t.join()

    (event,) = events

    (exception,) = event["exception"]["values"]

    assert exception["type"] == "ZeroDivisionError"
    assert exception["mechanism"]["type"] == "threading"
    assert not exception["mechanism"]["handled"]

    # Free-threaded builds set thread_inherit_context to True, otherwise thread_inherit_context is False
    if propagate_hub or getattr(sys.flags, "thread_inherit_context", None):
        assert event["tags"]["stage1"] == "true"
    else:
        assert "stage1" not in event.get("tags", {})


@pytest.mark.parametrize("propagate_hub", (True, False))
def test_propagates_threadpool_hub(sentry_init, capture_events, propagate_hub):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[ThreadingIntegration(propagate_hub=propagate_hub)],
    )
    events = capture_events()

    def double(number):
        with sentry_sdk.start_span(op="task", name=str(number)):
            return number * 2

    with sentry_sdk.start_transaction(name="test_handles_threadpool"):
        with futures.ThreadPoolExecutor(max_workers=1) as executor:
            tasks = [executor.submit(double, number) for number in [1, 2, 3, 4]]
            for future in futures.as_completed(tasks):
                print("Getting future value!", future.result())

    sentry_sdk.flush()

    # Free-threaded builds set thread_inherit_context to True, otherwise thread_inherit_context is False
    if propagate_hub or getattr(sys.flags, "thread_inherit_context", None):
        assert len(events) == 1
        (event,) = events
        assert event["spans"][0]["trace_id"] == event["spans"][1]["trace_id"]
        assert event["spans"][1]["trace_id"] == event["spans"][2]["trace_id"]
        assert event["spans"][2]["trace_id"] == event["spans"][3]["trace_id"]
        assert event["spans"][3]["trace_id"] == event["spans"][0]["trace_id"]
    else:
        (event,) = events
        assert len(event["spans"]) == 0


@pytest.mark.skip(reason="Temporarily disable to release SDK 2.0a1.")
def test_circular_references(sentry_init, request):
    sentry_init(default_integrations=False, integrations=[ThreadingIntegration()])

    gc.collect()
    gc.disable()
    request.addfinalizer(gc.enable)

    class MyThread(Thread):
        def run(self):
            pass

    t = MyThread()
    t.start()
    t.join()
    del t

    unreachable_objects = gc.collect()
    assert unreachable_objects == 0


@pytest.mark.filterwarnings("ignore:.*:pytest.PytestUnhandledThreadExceptionWarning")
def test_double_patching(sentry_init, capture_events):
    sentry_init(default_integrations=False, integrations=[ThreadingIntegration()])
    events = capture_events()

    # XXX: Workaround for race condition in the py library's magic import
    # system (py is a dependency of pytest)
    capture_message("hi")
    del events[:]

    class MyThread(Thread):
        def run(self):
            1 / 0

    ts = []
    for _ in range(10):
        t = MyThread()
        t.start()
        ts.append(t)

    for t in ts:
        t.join()

    assert len(events) == 10
    for event in events:
        (exception,) = event["exception"]["values"]
        assert exception["type"] == "ZeroDivisionError"


def test_wrapper_attributes(sentry_init):
    sentry_init(default_integrations=False, integrations=[ThreadingIntegration()])

    def target():
        assert t.run.__name__ == "run"
        assert t.run.__qualname__ == original_run.__qualname__

    t = Thread(target=target)
    t.start()
    t.join()

    assert Thread.start.__name__ == "start"
    assert Thread.start.__qualname__ == original_start.__qualname__
    assert t.start.__name__ == "start"
    assert t.start.__qualname__ == original_start.__qualname__

    assert Thread.run.__name__ == "run"
    assert Thread.run.__qualname__ == original_run.__qualname__
    assert t.run.__name__ == "run"
    assert t.run.__qualname__ == original_run.__qualname__


@pytest.mark.parametrize(
    "propagate_scope",
    (True, False),
    ids=["propagate_scope=True", "propagate_scope=False"],
)
def test_scope_data_not_leaked_in_threads(sentry_init, propagate_scope):
    sentry_init(
        integrations=[ThreadingIntegration(propagate_scope=propagate_scope)],
    )

    sentry_sdk.set_tag("initial_tag", "initial_value")
    initial_iso_scope = sentry_sdk.get_isolation_scope()

    def do_some_work():
        # check if we have the initial scope data propagated into the thread
        if propagate_scope:
            assert sentry_sdk.get_isolation_scope()._tags == {
                "initial_tag": "initial_value"
            }
        else:
            assert sentry_sdk.get_isolation_scope()._tags == {}

        # change data in isolation scope in thread
        sentry_sdk.set_tag("thread_tag", "thread_value")

    t = Thread(target=do_some_work)
    t.start()
    t.join()

    # check if the initial scope data is not modified by the started thread
    assert initial_iso_scope._tags == {"initial_tag": "initial_value"}, (
        "The isolation scope in the main thread should not be modified by the started thread."
    )


@pytest.mark.parametrize(
    "propagate_scope",
    (True, False),
    ids=["propagate_scope=True", "propagate_scope=False"],
)
def test_spans_from_multiple_threads(
    sentry_init, capture_events, render_span_tree, propagate_scope
):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[ThreadingIntegration(propagate_scope=propagate_scope)],
    )
    events = capture_events()

    def do_some_work(number):
        with sentry_sdk.start_span(
            op=f"inner-run-{number}", name=f"Thread: child-{number}"
        ):
            pass

    threads = []

    with sentry_sdk.start_transaction(op="outer-trx"):
        for number in range(5):
            with sentry_sdk.start_span(
                op=f"outer-submit-{number}", name="Thread: main"
            ):
                t = Thread(target=do_some_work, args=(number,))
                t.start()
                threads.append(t)

        for t in threads:
            t.join()

    (event,) = events

    # Free-threaded builds set thread_inherit_context to True, otherwise thread_inherit_context is False
    if propagate_scope or getattr(sys.flags, "thread_inherit_context", None):
        assert event["type"] == "transaction"
        assert render_span_tree(event["spans"], event["contexts"]["trace"]) == dedent(
            """\
            - op="outer-trx": description=null
              - op="outer-submit-0": description="Thread: main"
                - op="inner-run-0": description="Thread: child-0"
              - op="outer-submit-1": description="Thread: main"
                - op="inner-run-1": description="Thread: child-1"
              - op="outer-submit-2": description="Thread: main"
                - op="inner-run-2": description="Thread: child-2"
              - op="outer-submit-3": description="Thread: main"
                - op="inner-run-3": description="Thread: child-3"
              - op="outer-submit-4": description="Thread: main"
                - op="inner-run-4": description="Thread: child-4"\
"""
        )

    elif not propagate_scope:
        assert event["type"] == "transaction"
        assert render_span_tree(event["spans"], event["contexts"]["trace"]) == dedent(
            """\
            - op="outer-trx": description=null
              - op="outer-submit-0": description="Thread: main"
              - op="outer-submit-1": description="Thread: main"
              - op="outer-submit-2": description="Thread: main"
              - op="outer-submit-3": description="Thread: main"
              - op="outer-submit-4": description="Thread: main"\
"""
        )


@pytest.mark.parametrize(
    "propagate_scope",
    (True, False),
    ids=["propagate_scope=True", "propagate_scope=False"],
)
def test_spans_from_threadpool(
    sentry_init, capture_events, render_span_tree, propagate_scope
):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[ThreadingIntegration(propagate_scope=propagate_scope)],
    )
    events = capture_events()

    def do_some_work(number):
        with sentry_sdk.start_span(
            op=f"inner-run-{number}", name=f"Thread: child-{number}"
        ):
            pass

    with sentry_sdk.start_transaction(op="outer-trx"):
        with futures.ThreadPoolExecutor(max_workers=1) as executor:
            for number in range(5):
                with sentry_sdk.start_span(
                    op=f"outer-submit-{number}", name="Thread: main"
                ):
                    future = executor.submit(do_some_work, number)
                    future.result()

    (event,) = events

    # Free-threaded builds set thread_inherit_context to True, otherwise thread_inherit_context is False
    if propagate_scope or getattr(sys.flags, "thread_inherit_context", None):
        assert event["type"] == "transaction"
        assert render_span_tree(event["spans"], event["contexts"]["trace"]) == dedent(
            """\
            - op="outer-trx": description=null
              - op="outer-submit-0": description="Thread: main"
                - op="inner-run-0": description="Thread: child-0"
              - op="outer-submit-1": description="Thread: main"
                - op="inner-run-1": description="Thread: child-1"
              - op="outer-submit-2": description="Thread: main"
                - op="inner-run-2": description="Thread: child-2"
              - op="outer-submit-3": description="Thread: main"
                - op="inner-run-3": description="Thread: child-3"
              - op="outer-submit-4": description="Thread: main"
                - op="inner-run-4": description="Thread: child-4"\
"""
        )

    elif not propagate_scope:
        assert event["type"] == "transaction"
        assert render_span_tree(event["spans"], event["contexts"]["trace"]) == dedent(
            """\
            - op="outer-trx": description=null
              - op="outer-submit-0": description="Thread: main"
              - op="outer-submit-1": description="Thread: main"
              - op="outer-submit-2": description="Thread: main"
              - op="outer-submit-3": description="Thread: main"
              - op="outer-submit-4": description="Thread: main"\
"""
        )



import socket

import pytest

import sentry_sdk
from sentry_sdk import start_transaction
from sentry_sdk.integrations.socket import SocketIntegration
from tests.conftest import ApproxDict, create_mock_http_server

PORT = create_mock_http_server()


@pytest.mark.parametrize("span_streaming", [True, False])
def test_getaddrinfo_trace(sentry_init, capture_events, capture_items, span_streaming):
    sentry_init(
        integrations=[SocketIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="root"):
            socket.getaddrinfo("localhost", PORT)
        sentry_sdk.flush()

        spans = [item.payload for item in items if item.type == "span"]
        dns_span, _root = spans

        assert dns_span["attributes"]["sentry.op"] == "socket.dns"
        assert dns_span["attributes"]["sentry.origin"] == "auto.socket.socket"
        assert dns_span["name"] == f"localhost:{PORT}"  # noqa: E231
        assert dns_span["attributes"]["server.address"] == "localhost"
        assert dns_span["attributes"]["server.port"] == PORT
    else:
        events = capture_events()

        with start_transaction():
            socket.getaddrinfo("localhost", PORT)

        (event,) = events
        (span,) = event["spans"]

        assert span["op"] == "socket.dns"
        assert span["description"] == f"localhost:{PORT}"  # noqa: E231
        assert span["data"] == ApproxDict(
            {
                "host": "localhost",
                "port": PORT,
            }
        )


@pytest.mark.parametrize("span_streaming", [True, False])
def test_create_connection_trace(
    sentry_init, capture_events, capture_items, span_streaming
):
    timeout = 10

    sentry_init(
        integrations=[SocketIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="root"):
            socket.create_connection(("localhost", PORT), timeout, None)
        sentry_sdk.flush()

        spans = [item.payload for item in items if item.type == "span"]
        # as getaddrinfo gets called in create_connection it should also contain a dns span
        # spans finish in order: dns (inner) ends first, connect ends, then root
        dns_span, connect_span, _root = spans

        assert connect_span["attributes"]["sentry.op"] == "socket.connection"
        assert connect_span["name"] == f"localhost:{PORT}"  # noqa: E231
        assert connect_span["attributes"]["server.address"] == "localhost"
        assert connect_span["attributes"]["server.port"] == PORT

        assert dns_span["attributes"]["sentry.op"] == "socket.dns"
        assert dns_span["name"] == f"localhost:{PORT}"  # noqa: E231
        assert dns_span["attributes"]["server.address"] == "localhost"
        assert dns_span["attributes"]["server.port"] == PORT
    else:
        events = capture_events()

        with start_transaction():
            socket.create_connection(("localhost", PORT), timeout, None)

        (event,) = events
        (connect_span, dns_span) = event["spans"]
        # as getaddrinfo gets called in create_connection it should also contain a dns span

        assert connect_span["op"] == "socket.connection"
        assert connect_span["description"] == f"localhost:{PORT}"  # noqa: E231
        assert connect_span["data"] == ApproxDict(
            {
                "address": ["localhost", PORT],
                "timeout": timeout,
                "source_address": None,
            }
        )

        assert dns_span["op"] == "socket.dns"
        assert dns_span["description"] == f"localhost:{PORT}"  # noqa: E231
        assert dns_span["data"] == ApproxDict(
            {
                "host": "localhost",
                "port": PORT,
            }
        )


@pytest.mark.parametrize("span_streaming", [True, False])
def test_span_origin(sentry_init, capture_events, capture_items, span_streaming):
    sentry_init(
        integrations=[SocketIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="foo"):
            socket.create_connection(("localhost", PORT), 1, None)
        sentry_sdk.flush()

        spans = [item.payload for item in items if item.type == "span"]
        dns_span, connect_span, _root = spans

        assert connect_span["attributes"]["sentry.op"] == "socket.connection"
        assert connect_span["attributes"]["sentry.origin"] == "auto.socket.socket"

        assert dns_span["attributes"]["sentry.op"] == "socket.dns"
        assert dns_span["attributes"]["sentry.origin"] == "auto.socket.socket"
    else:
        events = capture_events()

        with start_transaction(name="foo"):
            socket.create_connection(("localhost", PORT), 1, None)

        (event,) = events

        assert event["contexts"]["trace"]["origin"] == "manual"

        assert event["spans"][0]["op"] == "socket.connection"
        assert event["spans"][0]["origin"] == "auto.socket.socket"

        assert event["spans"][1]["op"] == "socket.dns"
        assert event["spans"][1]["origin"] == "auto.socket.socket"



import uuid

import fakeredis
import pytest
from fakeredis import FakeStrictRedis

import sentry_sdk
from sentry_sdk.consts import SPANDATA
from sentry_sdk.integrations.redis import RedisIntegration
from sentry_sdk.integrations.redis.utils import _get_safe_key, _key_as_string
from sentry_sdk.utils import parse_version

FAKEREDIS_VERSION = parse_version(fakeredis.__version__)


@pytest.mark.parametrize("span_streaming", [True, False])
def test_no_cache_basic(sentry_init, capture_events, capture_items, span_streaming):
    sentry_init(
        integrations=[
            RedisIntegration(),
        ],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    connection = FakeStrictRedis()

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            connection.get("mycachekey")
        sentry_sdk.flush()

        assert len(items) == 2
        db_span, parent_span = items[0].payload, items[1].payload
        assert parent_span["name"] == "custom parent"
        assert db_span["attributes"]["sentry.op"] == "db.redis"
    else:
        events = capture_events()
        with sentry_sdk.start_transaction():
            connection.get("mycachekey")

        (event,) = events
        spans = event["spans"]
        assert len(spans) == 1
        assert spans[0]["op"] == "db.redis"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_cache_basic(sentry_init, capture_events, capture_items, span_streaming):
    sentry_init(
        integrations=[
            RedisIntegration(
                cache_prefixes=["mycache"],
            ),
        ],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    connection = FakeStrictRedis()

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            connection.hget("mycachekey", "myfield")
            connection.get("mycachekey")
            connection.set("mycachekey1", "bla")
            connection.setex("mycachekey2", 10, "blub")
            connection.mget("mycachekey1", "mycachekey2")
        sentry_sdk.flush()

        # Close order: db spans close before their wrapping cache span,
        # and the "custom parent" segment closes last.
        assert len(items) == 10
        payloads = [item.payload for item in items]

        # hget: db only (HGET is not a cache command)
        assert payloads[0]["attributes"]["sentry.op"] == "db.redis"
        assert payloads[0]["attributes"][SPANDATA.DB_OPERATION_NAME] == "HGET"

        # get: db then cache.get
        assert payloads[1]["attributes"]["sentry.op"] == "db.redis"
        assert payloads[1]["attributes"][SPANDATA.DB_OPERATION_NAME] == "GET"
        assert payloads[2]["attributes"]["sentry.op"] == "cache.get"
        assert payloads[2]["attributes"][SPANDATA.DB_QUERY_TEXT] == "GET 'mycachekey'"

        # set: db then cache.put
        assert payloads[3]["attributes"]["sentry.op"] == "db.redis"
        assert payloads[3]["attributes"][SPANDATA.DB_OPERATION_NAME] == "SET"
        assert payloads[4]["attributes"]["sentry.op"] == "cache.put"
        assert (
            payloads[4]["attributes"][SPANDATA.DB_QUERY_TEXT]
            == "SET 'mycachekey1' [Filtered]"
        )

        # setex: db then cache.put
        assert payloads[5]["attributes"]["sentry.op"] == "db.redis"
        assert payloads[5]["attributes"][SPANDATA.DB_OPERATION_NAME] == "SETEX"
        assert payloads[6]["attributes"]["sentry.op"] == "cache.put"
        assert (
            payloads[6]["attributes"][SPANDATA.DB_QUERY_TEXT]
            == "SETEX 'mycachekey2' [Filtered] [Filtered]"
        )

        # mget: db then cache.get
        assert payloads[7]["attributes"]["sentry.op"] == "db.redis"
        assert payloads[7]["attributes"][SPANDATA.DB_OPERATION_NAME] == "MGET"
        assert payloads[8]["attributes"]["sentry.op"] == "cache.get"
        assert (
            payloads[8]["attributes"][SPANDATA.DB_QUERY_TEXT]
            == "MGET 'mycachekey1' [Filtered]"
        )

        assert payloads[9]["name"] == "custom parent"
    else:
        events = capture_events()
        with sentry_sdk.start_transaction():
            connection.hget("mycachekey", "myfield")
            connection.get("mycachekey")
            connection.set("mycachekey1", "bla")
            connection.setex("mycachekey2", 10, "blub")
            connection.mget("mycachekey1", "mycachekey2")

        (event,) = events
        spans = event["spans"]
        assert len(spans) == 9

        # no cache support for hget command
        assert spans[0]["op"] == "db.redis"
        assert spans[0]["tags"]["redis.command"] == "HGET"

        assert spans[1]["op"] == "cache.get"
        assert spans[2]["op"] == "db.redis"
        assert spans[2]["tags"]["redis.command"] == "GET"

        assert spans[3]["op"] == "cache.put"
        assert spans[4]["op"] == "db.redis"
        assert spans[4]["tags"]["redis.command"] == "SET"

        assert spans[5]["op"] == "cache.put"
        assert spans[6]["op"] == "db.redis"
        assert spans[6]["tags"]["redis.command"] == "SETEX"

        assert spans[7]["op"] == "cache.get"
        assert spans[8]["op"] == "db.redis"
        assert spans[8]["tags"]["redis.command"] == "MGET"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_cache_keys(sentry_init, capture_events, capture_items, span_streaming):
    sentry_init(
        integrations=[
            RedisIntegration(
                cache_prefixes=["bla", "blub"],
            ),
        ],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    connection = FakeStrictRedis()

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            connection.get("somethingelse")
            connection.get("blub")
            connection.get("blubkeything")
            connection.get("bl")
        sentry_sdk.flush()

        assert len(items) == 7
        payloads = [item.payload for item in items]

        # somethingelse: db only
        assert payloads[0]["attributes"]["sentry.op"] == "db.redis"
        assert payloads[0]["name"] == "GET 'somethingelse'"

        # blub: db then cache.get
        assert payloads[1]["attributes"]["sentry.op"] == "db.redis"
        assert payloads[1]["name"] == "GET 'blub'"
        assert payloads[2]["attributes"]["sentry.op"] == "cache.get"
        assert payloads[2]["name"] == "blub"
        assert payloads[2]["attributes"][SPANDATA.DB_QUERY_TEXT] == "GET 'blub'"

        # blubkeything: db then cache.get
        assert payloads[3]["attributes"]["sentry.op"] == "db.redis"
        assert payloads[3]["name"] == "GET 'blubkeything'"
        assert payloads[4]["attributes"]["sentry.op"] == "cache.get"
        assert payloads[4]["name"] == "blubkeything"
        assert payloads[4]["attributes"][SPANDATA.DB_QUERY_TEXT] == "GET 'blubkeything'"

        # bl: db only (no prefix match)
        assert payloads[5]["attributes"]["sentry.op"] == "db.redis"
        assert payloads[5]["name"] == "GET 'bl'"

        assert payloads[6]["name"] == "custom parent"
    else:
        events = capture_events()
        with sentry_sdk.start_transaction():
            connection.get("somethingelse")
            connection.get("blub")
            connection.get("blubkeything")
            connection.get("bl")

        (event,) = events
        spans = event["spans"]
        assert len(spans) == 6
        assert spans[0]["op"] == "db.redis"
        assert spans[0]["description"] == "GET 'somethingelse'"

        assert spans[1]["op"] == "cache.get"
        assert spans[1]["description"] == "blub"
        assert spans[2]["op"] == "db.redis"
        assert spans[2]["description"] == "GET 'blub'"

        assert spans[3]["op"] == "cache.get"
        assert spans[3]["description"] == "blubkeything"
        assert spans[4]["op"] == "db.redis"
        assert spans[4]["description"] == "GET 'blubkeything'"

        assert spans[5]["op"] == "db.redis"
        assert spans[5]["description"] == "GET 'bl'"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_cache_data(sentry_init, capture_events, capture_items, span_streaming):
    sentry_init(
        integrations=[
            RedisIntegration(
                cache_prefixes=["mycache"],
            ),
        ],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    # Use a unique host per parametrized run so fakeredis (which shares state
    # keyed by host:port) doesn't leak the SET from a prior run into this one.
    host = f"mycacheserver-{uuid.uuid4().hex}.io"
    connection = FakeStrictRedis(host=host, port=6378)

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            connection.get("mycachekey")
            connection.set("mycachekey", "事实胜于雄辩")
            connection.get("mycachekey")
        sentry_sdk.flush()

        # Close order: db then cache for each command, then parent
        assert len(items) == 7
        payloads = [item.payload for item in items]

        # First get (miss)
        assert payloads[0]["attributes"]["sentry.op"] == "db.redis"
        cache_get_miss = payloads[1]
        assert cache_get_miss["attributes"]["sentry.op"] == "cache.get"
        assert cache_get_miss["name"] == "mycachekey"
        assert cache_get_miss["attributes"]["cache.key"] == ["mycachekey"]
        assert cache_get_miss["attributes"]["cache.hit"] is False
        assert "cache.item_size" not in cache_get_miss["attributes"]
        if FAKEREDIS_VERSION <= (2, 7, 1):
            assert "network.peer.port" not in cache_get_miss["attributes"]
        else:
            assert cache_get_miss["attributes"]["network.peer.port"] == 6378
        if FAKEREDIS_VERSION <= (1, 7, 1):
            assert "network.peer.address" not in cache_get_miss["attributes"]
        else:
            assert cache_get_miss["attributes"]["network.peer.address"] == host

        # Set
        assert payloads[2]["attributes"]["sentry.op"] == "db.redis"
        cache_put = payloads[3]
        assert cache_put["attributes"]["sentry.op"] == "cache.put"
        assert cache_put["name"] == "mycachekey"
        assert cache_put["attributes"]["cache.key"] == ["mycachekey"]
        assert "cache.hit" not in cache_put["attributes"]
        assert cache_put["attributes"]["cache.item_size"] == 18
        if FAKEREDIS_VERSION <= (2, 7, 1):
            assert "network.peer.port" not in cache_put["attributes"]
        else:
            assert cache_put["attributes"]["network.peer.port"] == 6378
        if FAKEREDIS_VERSION <= (1, 7, 1):
            assert "network.peer.address" not in cache_put["attributes"]
        else:
            assert cache_put["attributes"]["network.peer.address"] == host

        # Second get (hit)
        assert payloads[4]["attributes"]["sentry.op"] == "db.redis"
        cache_get_hit = payloads[5]
        assert cache_get_hit["attributes"]["sentry.op"] == "cache.get"
        assert cache_get_hit["attributes"]["cache.key"] == ["mycachekey"]
        assert cache_get_hit["attributes"]["cache.hit"] is True
        assert cache_get_hit["attributes"]["cache.item_size"] == 18
        if FAKEREDIS_VERSION <= (2, 7, 1):
            assert "network.peer.port" not in cache_get_hit["attributes"]
        else:
            assert cache_get_hit["attributes"]["network.peer.port"] == 6378
        if FAKEREDIS_VERSION <= (1, 7, 1):
            assert "network.peer.address" not in cache_get_hit["attributes"]
        else:
            assert cache_get_hit["attributes"]["network.peer.address"] == host

        assert payloads[6]["name"] == "custom parent"
    else:
        events = capture_events()
        with sentry_sdk.start_transaction():
            connection.get("mycachekey")
            connection.set("mycachekey", "事实胜于雄辩")
            connection.get("mycachekey")

        (event,) = events
        spans = event["spans"]

        assert len(spans) == 6

        assert spans[0]["op"] == "cache.get"
        assert spans[0]["description"] == "mycachekey"
        assert spans[0]["data"]["cache.key"] == [
            "mycachekey",
        ]
        assert spans[0]["data"]["cache.hit"] == False  # noqa: E712
        assert "cache.item_size" not in spans[0]["data"]
        # very old fakeredis can not handle port and/or host.
        # only applicable for Redis v3
        if FAKEREDIS_VERSION <= (2, 7, 1):
            assert "network.peer.port" not in spans[0]["data"]
        else:
            assert spans[0]["data"]["network.peer.port"] == 6378
        if FAKEREDIS_VERSION <= (1, 7, 1):
            assert "network.peer.address" not in spans[0]["data"]
        else:
            assert spans[0]["data"]["network.peer.address"] == host

        assert spans[1]["op"] == "db.redis"  # we ignore db spans in this test.

        assert spans[2]["op"] == "cache.put"
        assert spans[2]["description"] == "mycachekey"
        assert spans[2]["data"]["cache.key"] == [
            "mycachekey",
        ]
        assert "cache.hit" not in spans[1]["data"]
        assert spans[2]["data"]["cache.item_size"] == 18
        # very old fakeredis can not handle port.
        # only used with redis v3
        if FAKEREDIS_VERSION <= (2, 7, 1):
            assert "network.peer.port" not in spans[2]["data"]
        else:
            assert spans[2]["data"]["network.peer.port"] == 6378
        if FAKEREDIS_VERSION <= (1, 7, 1):
            assert "network.peer.address" not in spans[2]["data"]
        else:
            assert spans[2]["data"]["network.peer.address"] == host

        assert spans[3]["op"] == "db.redis"  # we ignore db spans in this test.

        assert spans[4]["op"] == "cache.get"
        assert spans[4]["description"] == "mycachekey"
        assert spans[4]["data"]["cache.key"] == [
            "mycachekey",
        ]
        assert spans[4]["data"]["cache.hit"] == True  # noqa: E712
        assert spans[4]["data"]["cache.item_size"] == 18
        # very old fakeredis can not handle port.
        # only used with redis v3
        if FAKEREDIS_VERSION <= (2, 7, 1):
            assert "network.peer.port" not in spans[4]["data"]
        else:
            assert spans[4]["data"]["network.peer.port"] == 6378
        if FAKEREDIS_VERSION <= (1, 7, 1):
            assert "network.peer.address" not in spans[4]["data"]
        else:
            assert spans[4]["data"]["network.peer.address"] == host

        assert spans[5]["op"] == "db.redis"  # we ignore db spans in this test.


@pytest.mark.parametrize("span_streaming", [True, False])
def test_cache_prefixes(sentry_init, capture_events, capture_items, span_streaming):
    sentry_init(
        integrations=[
            RedisIntegration(
                cache_prefixes=["yes"],
            ),
        ],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    connection = FakeStrictRedis()

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            connection.mget("yes", "no")
            connection.mget("no", 1, "yes")
            connection.mget("no", "yes.1", "yes.2")
            connection.mget("no.1", "no.2", "no.3")
            connection.mget("no.1", "no.2", "no.actually.yes")
            connection.mget(b"no.3", b"yes.5")
            connection.mget(uuid.uuid4().bytes)
            connection.mget(uuid.uuid4().bytes, "yes")
        sentry_sdk.flush()

        # 8 db spans + 5 cache spans + 1 parent
        assert len(items) == 14
        payloads = [item.payload for item in items]
        assert payloads[-1]["name"] == "custom parent"

        cache_spans = [
            p for p in payloads if p["attributes"].get("sentry.op") == "cache.get"
        ]
        assert len(cache_spans) == 5

        assert cache_spans[0]["name"] == "yes, no"
        assert cache_spans[1]["name"] == "no, 1, yes"
        assert cache_spans[2]["name"] == "no, yes.1, yes.2"
        assert cache_spans[3]["name"] == "no.3, yes.5"
        assert cache_spans[4]["name"] == ", yes"
    else:
        events = capture_events()
        with sentry_sdk.start_transaction():
            connection.mget("yes", "no")
            connection.mget("no", 1, "yes")
            connection.mget("no", "yes.1", "yes.2")
            connection.mget("no.1", "no.2", "no.3")
            connection.mget("no.1", "no.2", "no.actually.yes")
            connection.mget(b"no.3", b"yes.5")
            connection.mget(uuid.uuid4().bytes)
            connection.mget(uuid.uuid4().bytes, "yes")

        (event,) = events

        spans = event["spans"]
        assert len(spans) == 13  # 8 db spans + 5 cache spans

        cache_spans = [span for span in spans if span["op"] == "cache.get"]
        assert len(cache_spans) == 5

        assert cache_spans[0]["description"] == "yes, no"
        assert cache_spans[1]["description"] == "no, 1, yes"
        assert cache_spans[2]["description"] == "no, yes.1, yes.2"
        assert cache_spans[3]["description"] == "no.3, yes.5"
        assert cache_spans[4]["description"] == ", yes"


@pytest.mark.parametrize(
    "method_name,args,kwargs,expected_key",
    [
        (None, None, None, None),
        ("", None, None, None),
        ("set", ["bla", "valuebla"], None, ("bla",)),
        ("setex", ["bla", 10, "valuebla"], None, ("bla",)),
        ("get", ["bla"], None, ("bla",)),
        ("mget", ["bla", "blub", "foo"], None, ("bla", "blub", "foo")),
        ("set", [b"bla", "valuebla"], None, (b"bla",)),
        ("setex", [b"bla", 10, "valuebla"], None, (b"bla",)),
        ("get", [b"bla"], None, (b"bla",)),
        ("mget", [b"bla", "blub", "foo"], None, (b"bla", "blub", "foo")),
        ("not-important", None, {"something": "bla"}, None),
        ("not-important", None, {"key": None}, None),
        ("not-important", None, {"key": "bla"}, ("bla",)),
        ("not-important", None, {"key": b"bla"}, (b"bla",)),
        ("not-important", None, {"key": []}, None),
        (
            "not-important",
            None,
            {
                "key": [
                    "bla",
                ]
            },
            ("bla",),
        ),
        (
            "not-important",
            None,
            {"key": [b"bla", "blub", "foo"]},
            (b"bla", "blub", "foo"),
        ),
        (
            "not-important",
            None,
            {"key": b"\x00c\x0f\xeaC\xe1L\x1c\xbff\xcb\xcc\xc1\xed\xc6\t"},
            (b"\x00c\x0f\xeaC\xe1L\x1c\xbff\xcb\xcc\xc1\xed\xc6\t",),
        ),
        (
            "get",
            [b"\x00c\x0f\xeaC\xe1L\x1c\xbff\xcb\xcc\xc1\xed\xc6\t"],
            None,
            (b"\x00c\x0f\xeaC\xe1L\x1c\xbff\xcb\xcc\xc1\xed\xc6\t",),
        ),
        (
            "get",
            [123],
            None,
            (123,),
        ),
    ],
)
def test_get_safe_key(method_name, args, kwargs, expected_key):
    assert _get_safe_key(method_name, args, kwargs) == expected_key


@pytest.mark.parametrize(
    "key,expected_key",
    [
        (None, ""),
        (("bla",), "bla"),
        (("bla", "blub", "foo"), "bla, blub, foo"),
        ((b"bla",), "bla"),
        ((b"bla", "blub", "foo"), "bla, blub, foo"),
        (
            [
                "bla",
            ],
            "bla",
        ),
        (["bla", "blub", "foo"], "bla, blub, foo"),
        ([uuid.uuid4().bytes], ""),
        ({"key1": 1, "key2": 2}, "key1, key2"),
        (1, "1"),
        ([1, 2, 3, b"hello"], "1, 2, 3, hello"),
    ],
)
def test_key_as_string(key, expected_key):
    assert _key_as_string(key) == expected_key



try:
    from importlib.metadata import entry_points
except ImportError:
    from importlib_metadata import entry_points

from sentry_sdk.integrations.opentelemetry import SentryPropagator


def test_propagator_registered_as_entry_point():
    all_eps = entry_points()

    if isinstance(all_eps, dict):
        # Python 3.7-3.8 stdlib returns a dict keyed by group
        matches = [
            ep
            for ep in all_eps.get("opentelemetry_propagator", [])
            if ep.name == "sentry"
        ]
    else:
        matches = list(all_eps.select(group="opentelemetry_propagator", name="sentry"))

    assert len(matches) >= 1
    assert matches[0].value == "sentry_sdk.integrations.opentelemetry:SentryPropagator"

    loaded = matches[0].load()
    assert loaded is SentryPropagator



import concurrent.futures as cf
import sys

import ldclient
import pytest
from ldclient import LDClient
from ldclient.config import Config
from ldclient.context import Context
from ldclient.integrations.test_data import TestData

import sentry_sdk
from sentry_sdk import start_span, start_transaction
from sentry_sdk.integrations import DidNotEnable
from sentry_sdk.integrations.launchdarkly import LaunchDarklyIntegration
from tests.conftest import ApproxDict


@pytest.mark.parametrize(
    "use_global_client",
    (False, True),
)
def test_launchdarkly_integration(
    sentry_init, use_global_client, capture_events, uninstall_integration
):
    td = TestData.data_source()
    td.update(td.flag("hello").variation_for_all(True))
    td.update(td.flag("world").variation_for_all(True))
    # Disable background requests as we aren't using a server.
    config = Config(
        "sdk-key", update_processor_class=td, diagnostic_opt_out=True, send_events=False
    )

    uninstall_integration(LaunchDarklyIntegration.identifier)
    if use_global_client:
        ldclient.set_config(config)
        sentry_init(integrations=[LaunchDarklyIntegration()])
        client = ldclient.get()
    else:
        client = LDClient(config=config)
        sentry_init(integrations=[LaunchDarklyIntegration(ld_client=client)])

    # Evaluate
    client.variation("hello", Context.create("my-org", "organization"), False)
    client.variation("world", Context.create("user1", "user"), False)
    client.variation("other", Context.create("user2", "user"), False)

    events = capture_events()
    sentry_sdk.capture_exception(Exception("something wrong!"))

    assert len(events) == 1
    assert events[0]["contexts"]["flags"] == {
        "values": [
            {"flag": "hello", "result": True},
            {"flag": "world", "result": True},
            {"flag": "other", "result": False},
        ]
    }


def test_launchdarkly_integration_threaded(
    sentry_init, capture_events, uninstall_integration
):
    td = TestData.data_source()
    td.update(td.flag("hello").variation_for_all(True))
    td.update(td.flag("world").variation_for_all(True))
    client = LDClient(
        config=Config(
            "sdk-key",
            update_processor_class=td,
            diagnostic_opt_out=True,  # Disable background requests as we aren't using a server.
            send_events=False,
        )
    )
    context = Context.create("user1")

    uninstall_integration(LaunchDarklyIntegration.identifier)
    sentry_init(integrations=[LaunchDarklyIntegration(ld_client=client)])
    events = capture_events()

    def task(flag_key):
        # Creates a new isolation scope for the thread.
        # This means the evaluations in each task are captured separately.
        with sentry_sdk.isolation_scope():
            client.variation(flag_key, context, False)
            # use a tag to identify to identify events later on
            sentry_sdk.set_tag("task_id", flag_key)
            sentry_sdk.capture_exception(Exception("something wrong!"))

    # Capture an eval before we split isolation scopes.
    client.variation("hello", context, False)

    with cf.ThreadPoolExecutor(max_workers=2) as pool:
        pool.map(task, ["world", "other"])

    # Capture error in original scope
    sentry_sdk.set_tag("task_id", "0")
    sentry_sdk.capture_exception(Exception("something wrong!"))

    assert len(events) == 3
    events.sort(key=lambda e: e["tags"]["task_id"])

    assert events[0]["contexts"]["flags"] == {
        "values": [
            {"flag": "hello", "result": True},
        ]
    }
    assert events[1]["contexts"]["flags"] == {
        "values": [
            {"flag": "hello", "result": True},
            {"flag": "other", "result": False},
        ]
    }
    assert events[2]["contexts"]["flags"] == {
        "values": [
            {"flag": "hello", "result": True},
            {"flag": "world", "result": True},
        ]
    }


@pytest.mark.skipif(sys.version_info < (3, 7), reason="requires python3.7 or higher")
def test_launchdarkly_integration_asyncio(
    sentry_init, capture_events, uninstall_integration
):
    """Assert concurrently evaluated flags do not pollute one another."""

    asyncio = pytest.importorskip("asyncio")

    td = TestData.data_source()
    td.update(td.flag("hello").variation_for_all(True))
    td.update(td.flag("world").variation_for_all(True))
    client = LDClient(
        config=Config(
            "sdk-key",
            update_processor_class=td,
            diagnostic_opt_out=True,  # Disable background requests as we aren't using a server.
            send_events=False,
        )
    )
    context = Context.create("user1")

    uninstall_integration(LaunchDarklyIntegration.identifier)
    sentry_init(integrations=[LaunchDarklyIntegration(ld_client=client)])
    events = capture_events()

    async def task(flag_key):
        with sentry_sdk.isolation_scope():
            client.variation(flag_key, context, False)
            # use a tag to identify to identify events later on
            sentry_sdk.set_tag("task_id", flag_key)
            sentry_sdk.capture_exception(Exception("something wrong!"))

    async def runner():
        return asyncio.gather(task("world"), task("other"))

    # Capture an eval before we split isolation scopes.
    client.variation("hello", context, False)

    asyncio.run(runner())

    # Capture error in original scope
    sentry_sdk.set_tag("task_id", "0")
    sentry_sdk.capture_exception(Exception("something wrong!"))

    assert len(events) == 3
    events.sort(key=lambda e: e["tags"]["task_id"])

    assert events[0]["contexts"]["flags"] == {
        "values": [
            {"flag": "hello", "result": True},
        ]
    }
    assert events[1]["contexts"]["flags"] == {
        "values": [
            {"flag": "hello", "result": True},
            {"flag": "other", "result": False},
        ]
    }
    assert events[2]["contexts"]["flags"] == {
        "values": [
            {"flag": "hello", "result": True},
            {"flag": "world", "result": True},
        ]
    }


def test_launchdarkly_integration_did_not_enable(monkeypatch):
    # Client is not passed in and set_config wasn't called.
    # TODO: Bad practice to access internals like this. We can skip this test, or remove this
    #  case entirely (force user to pass in a client instance).
    ldclient._reset_client()
    try:
        ldclient.__lock.lock()
        ldclient.__config = None
    finally:
        ldclient.__lock.unlock()

    with pytest.raises(DidNotEnable):
        LaunchDarklyIntegration()

    td = TestData.data_source()
    # Disable background requests as we aren't using a server.
    # Required because we corrupt the internal state above.
    config = Config(
        "sdk-key", update_processor_class=td, diagnostic_opt_out=True, send_events=False
    )
    # Client not initialized.
    client = LDClient(config=config)
    monkeypatch.setattr(client, "is_initialized", lambda: False)
    with pytest.raises(DidNotEnable):
        LaunchDarklyIntegration(ld_client=client)


@pytest.mark.parametrize(
    "use_global_client",
    (False, True),
)
@pytest.mark.parametrize(
    "span_streaming",
    [True, False],
)
def test_launchdarkly_span_integration(
    sentry_init,
    use_global_client,
    capture_events,
    capture_items,
    uninstall_integration,
    span_streaming,
):
    td = TestData.data_source()
    td.update(td.flag("hello").variation_for_all(True))
    # Disable background requests as we aren't using a server.
    config = Config(
        "sdk-key", update_processor_class=td, diagnostic_opt_out=True, send_events=False
    )

    uninstall_integration(LaunchDarklyIntegration.identifier)
    if use_global_client:
        ldclient.set_config(config)
        sentry_init(
            traces_sample_rate=1.0,
            integrations=[LaunchDarklyIntegration()],
            trace_lifecycle="stream" if span_streaming else "static",
        )
        client = ldclient.get()
    else:
        client = LDClient(config=config)
        sentry_init(
            traces_sample_rate=1.0,
            integrations=[LaunchDarklyIntegration(ld_client=client)],
            trace_lifecycle="stream" if span_streaming else "static",
        )

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="bar"):
            client.variation("hello", Context.create("my-org", "organization"), False)
            client.variation("other", Context.create("my-org", "organization"), False)

        sentry_sdk.flush()

        assert len(items) == 1
        span = items[0].payload
        assert span["attributes"]["flag.evaluation.hello"] is True
        assert span["attributes"]["flag.evaluation.other"] is False

    else:
        events = capture_events()

        with start_transaction(name="hi"):
            with start_span(op="foo", name="bar"):
                client.variation(
                    "hello", Context.create("my-org", "organization"), False
                )
                client.variation(
                    "other", Context.create("my-org", "organization"), False
                )

        (event,) = events
        assert event["spans"][0]["data"] == ApproxDict(
            {"flag.evaluation.hello": True, "flag.evaluation.other": False}
        )



from unittest import mock
from unittest.mock import MagicMock

import pytest
from opentelemetry import trace
from opentelemetry.context import get_current
from opentelemetry.trace import (
    SpanContext,
    TraceFlags,
    set_span_in_context,
)
from opentelemetry.trace.propagation import get_current_span

from sentry_sdk.integrations.opentelemetry.consts import (
    SENTRY_BAGGAGE_KEY,
    SENTRY_TRACE_KEY,
)
from sentry_sdk.integrations.opentelemetry.integration import (
    OpenTelemetryIntegration,
)
from sentry_sdk.integrations.opentelemetry.propagator import SentryPropagator
from sentry_sdk.integrations.opentelemetry.span_processor import SentrySpanProcessor
from sentry_sdk.scope import global_event_processors
from sentry_sdk.tracing_utils import Baggage


@pytest.mark.forked
def test_extract_no_context_no_sentry_trace_header():
    """
    No context and NO Sentry trace data in getter.
    Extract should return empty context.
    """
    carrier = None
    context = None
    getter = MagicMock()
    getter.get.return_value = None

    modified_context = SentryPropagator().extract(carrier, context, getter)

    assert modified_context == {}


@pytest.mark.forked
def test_extract_context_no_sentry_trace_header():
    """
    Context but NO Sentry trace data in getter.
    Extract should return context as is.
    """
    carrier = None
    context = {"some": "value"}
    getter = MagicMock()
    getter.get.return_value = None

    modified_context = SentryPropagator().extract(carrier, context, getter)

    assert modified_context == context


@pytest.mark.forked
def test_extract_empty_context_sentry_trace_header_no_baggage():
    """
    Empty context but Sentry trace data but NO Baggage in getter.
    Extract should return context that has empty baggage in it and also a NoopSpan with span_id and trace_id.
    """
    carrier = None
    context = {}
    getter = MagicMock()
    getter.get.side_effect = [
        ["1234567890abcdef1234567890abcdef-1234567890abcdef-1"],
        None,
    ]

    modified_context = SentryPropagator().extract(carrier, context, getter)

    assert len(modified_context.keys()) == 3

    assert modified_context[SENTRY_TRACE_KEY] == {
        "trace_id": "1234567890abcdef1234567890abcdef",
        "parent_span_id": "1234567890abcdef",
        "parent_sampled": True,
    }
    assert modified_context[SENTRY_BAGGAGE_KEY].serialize() == ""

    span_context = get_current_span(modified_context).get_span_context()
    assert span_context.span_id == int("1234567890abcdef", 16)
    assert span_context.trace_id == int("1234567890abcdef1234567890abcdef", 16)


@pytest.mark.forked
def test_extract_context_sentry_trace_header_baggage():
    """
    Empty context but Sentry trace data and Baggage in getter.
    Extract should return context that has baggage in it and also a NoopSpan with span_id and trace_id.
    """
    baggage_header = (
        "other-vendor-value-1=foo;bar;baz, sentry-trace_id=771a43a4192642f0b136d5159a501700, "
        "sentry-public_key=49d0f7386ad645858ae85020e393bef3, sentry-sample_rate=0.01337, "
        "sentry-user_id=Am%C3%A9lie, other-vendor-value-2=foo;bar;"
    )

    carrier = None
    context = {"some": "value"}
    getter = MagicMock()
    getter.get.side_effect = [
        ["1234567890abcdef1234567890abcdef-1234567890abcdef-1"],
        [baggage_header],
    ]

    modified_context = SentryPropagator().extract(carrier, context, getter)

    assert len(modified_context.keys()) == 4

    assert modified_context[SENTRY_TRACE_KEY] == {
        "trace_id": "1234567890abcdef1234567890abcdef",
        "parent_span_id": "1234567890abcdef",
        "parent_sampled": True,
    }

    assert modified_context[SENTRY_BAGGAGE_KEY].serialize() == (
        "sentry-trace_id=771a43a4192642f0b136d5159a501700,"
        "sentry-public_key=49d0f7386ad645858ae85020e393bef3,"
        "sentry-sample_rate=0.01337,sentry-user_id=Am%C3%A9lie"
    )

    span_context = get_current_span(modified_context).get_span_context()
    assert span_context.span_id == int("1234567890abcdef", 16)
    assert span_context.trace_id == int("1234567890abcdef1234567890abcdef", 16)


@pytest.mark.forked
def test_inject_empty_otel_span_map():
    """
    Empty otel_span_map.
    So there is no sentry_span to be found in inject()
    and the function is returned early and no setters are called.
    """
    carrier = None
    context = get_current()
    setter = MagicMock()
    setter.set = MagicMock()

    span_context = SpanContext(
        trace_id=int("1234567890abcdef1234567890abcdef", 16),
        span_id=int("1234567890abcdef", 16),
        trace_flags=TraceFlags(TraceFlags.SAMPLED),
        is_remote=True,
    )
    span = MagicMock()
    span.get_span_context.return_value = span_context

    with mock.patch(
        "sentry_sdk.integrations.opentelemetry.propagator.trace.get_current_span",
        return_value=span,
    ):
        full_context = set_span_in_context(span, context)
        SentryPropagator().inject(carrier, full_context, setter)

        setter.set.assert_not_called()


@pytest.mark.forked
def test_inject_sentry_span_no_baggage():
    """
    Inject a sentry span with no baggage.
    """
    carrier = None
    context = get_current()
    setter = MagicMock()
    setter.set = MagicMock()

    trace_id = "1234567890abcdef1234567890abcdef"
    span_id = "1234567890abcdef"

    span_context = SpanContext(
        trace_id=int(trace_id, 16),
        span_id=int(span_id, 16),
        trace_flags=TraceFlags(TraceFlags.SAMPLED),
        is_remote=True,
    )
    span = MagicMock()
    span.get_span_context.return_value = span_context

    sentry_span = MagicMock()
    sentry_span.to_traceparent = mock.Mock(
        return_value="1234567890abcdef1234567890abcdef-1234567890abcdef-1"
    )
    sentry_span.containing_transaction.get_baggage = mock.Mock(return_value=None)

    span_processor = SentrySpanProcessor()
    span_processor.otel_span_map[span_id] = sentry_span

    with mock.patch(
        "sentry_sdk.integrations.opentelemetry.propagator.trace.get_current_span",
        return_value=span,
    ):
        full_context = set_span_in_context(span, context)
        SentryPropagator().inject(carrier, full_context, setter)

        setter.set.assert_called_once_with(
            carrier,
            "sentry-trace",
            "1234567890abcdef1234567890abcdef-1234567890abcdef-1",
        )


def test_inject_sentry_span_empty_baggage():
    """
    Inject a sentry span with no baggage.
    """
    carrier = None
    context = get_current()
    setter = MagicMock()
    setter.set = MagicMock()

    trace_id = "1234567890abcdef1234567890abcdef"
    span_id = "1234567890abcdef"

    span_context = SpanContext(
        trace_id=int(trace_id, 16),
        span_id=int(span_id, 16),
        trace_flags=TraceFlags(TraceFlags.SAMPLED),
        is_remote=True,
    )
    span = MagicMock()
    span.get_span_context.return_value = span_context

    sentry_span = MagicMock()
    sentry_span.to_traceparent = mock.Mock(
        return_value="1234567890abcdef1234567890abcdef-1234567890abcdef-1"
    )
    sentry_span.containing_transaction.get_baggage = mock.Mock(return_value=Baggage({}))

    span_processor = SentrySpanProcessor()
    span_processor.otel_span_map[span_id] = sentry_span

    with mock.patch(
        "sentry_sdk.integrations.opentelemetry.propagator.trace.get_current_span",
        return_value=span,
    ):
        full_context = set_span_in_context(span, context)
        SentryPropagator().inject(carrier, full_context, setter)

        setter.set.assert_called_once_with(
            carrier,
            "sentry-trace",
            "1234567890abcdef1234567890abcdef-1234567890abcdef-1",
        )


def test_inject_sentry_span_baggage():
    """
    Inject a sentry span with baggage.
    """
    carrier = None
    context = get_current()
    setter = MagicMock()
    setter.set = MagicMock()

    trace_id = "1234567890abcdef1234567890abcdef"
    span_id = "1234567890abcdef"

    span_context = SpanContext(
        trace_id=int(trace_id, 16),
        span_id=int(span_id, 16),
        trace_flags=TraceFlags(TraceFlags.SAMPLED),
        is_remote=True,
    )
    span = MagicMock()
    span.get_span_context.return_value = span_context

    sentry_span = MagicMock()
    sentry_span.to_traceparent = mock.Mock(
        return_value="1234567890abcdef1234567890abcdef-1234567890abcdef-1"
    )
    sentry_items = {
        "sentry-trace_id": "771a43a4192642f0b136d5159a501700",
        "sentry-public_key": "49d0f7386ad645858ae85020e393bef3",
        "sentry-sample_rate": 0.01337,
        "sentry-user_id": "Amélie",
    }
    baggage = Baggage(sentry_items=sentry_items)
    sentry_span.containing_transaction.get_baggage = MagicMock(return_value=baggage)

    span_processor = SentrySpanProcessor()
    span_processor.otel_span_map[span_id] = sentry_span

    with mock.patch(
        "sentry_sdk.integrations.opentelemetry.propagator.trace.get_current_span",
        return_value=span,
    ):
        full_context = set_span_in_context(span, context)
        SentryPropagator().inject(carrier, full_context, setter)

        setter.set.assert_any_call(
            carrier,
            "sentry-trace",
            "1234567890abcdef1234567890abcdef-1234567890abcdef-1",
        )

        setter.set.assert_any_call(
            carrier,
            "baggage",
            baggage.serialize(),
        )


def test_inject_no_memory_leak():

    OpenTelemetryIntegration.setup_once()

    tracer = trace.get_tracer(__name__)
    propagator = SentryPropagator()

    cnt_before = len(global_event_processors)

    with tracer.start_as_current_span("bar") as new_span:
        context = set_span_in_context(new_span)
        carrier = "any_carrier"
        propagator.inject(carrier, context)

    cnt_after = len(global_event_processors)
    assert cnt_after == cnt_before



import itertools
import os
from datetime import datetime

import pytest
from django.contrib.auth.models import User
from django.db import connections

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse

from werkzeug.test import Client

import sentry_sdk
from sentry_sdk import start_transaction
from sentry_sdk.consts import SPANDATA, SPANNAME
from sentry_sdk.integrations.django import DjangoIntegration
from tests.integrations.django.myapp.wsgi import application
from tests.integrations.django.utils import pytest_mark_django_db_decorator


@pytest.fixture
def client():
    return Client(application)


@pytest.mark.forked
@pytest_mark_django_db_decorator(transaction=True)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_db_transaction_spans_disabled_no_autocommit(
    sentry_init,
    client,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[DjangoIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if "postgres" not in connections:
        pytest.skip("postgres tests disabled")

    # trigger Django to open a new connection by marking the existing one as None.
    connections["postgres"].connection = None

    if span_streaming:
        items = capture_items("span")

        client.get(reverse("postgres_insert_orm_no_autocommit_rollback"))
        client.get(reverse("postgres_insert_orm_no_autocommit"))

        with sentry_sdk.traces.start_span(name="custom parent"):
            from django.db import connection, transaction

            cursor = connection.cursor()

            query = """INSERT INTO auth_user (
    password,
    is_superuser,
    username,
    first_name,
    last_name,
    email,
    is_staff,
    is_active,
    date_joined
)
VALUES ('password', false, %s, %s, %s, %s, false, true, %s);"""

            query_list = (
                (
                    "user1",
                    "John",
                    "Doe",
                    "user1@example.com",
                    datetime(1970, 1, 1),
                ),
                (
                    "user2",
                    "Max",
                    "Mustermann",
                    "user2@example.com",
                    datetime(1970, 1, 1),
                ),
            )

            transaction.set_autocommit(False)
            cursor.executemany(query, query_list)
            transaction.rollback()
            transaction.set_autocommit(True)

        with sentry_sdk.traces.start_span(name="custom parent"):
            from django.db import connection, transaction

            cursor = connection.cursor()

            query = """INSERT INTO auth_user (
    password,
    is_superuser,
    username,
    first_name,
    last_name,
    email,
    is_staff,
    is_active,
    date_joined
)
VALUES ('password', false, %s, %s, %s, %s, false, true, %s);"""

            query_list = (
                (
                    "user1",
                    "John",
                    "Doe",
                    "user1@example.com",
                    datetime(1970, 1, 1),
                ),
                (
                    "user2",
                    "Max",
                    "Mustermann",
                    "user2@example.com",
                    datetime(1970, 1, 1),
                ),
            )

            transaction.set_autocommit(False)
            cursor.executemany(query, query_list)
            transaction.commit()
            transaction.set_autocommit(True)

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        postgres_rollback = spans[4]
        assert postgres_rollback["is_segment"] is True
        postgres_commit = spans[9]
        assert postgres_commit["is_segment"] is True
        sqlite_rollback = spans[11]
        assert sqlite_rollback["is_segment"] is True
        sqlite_commit = spans[13]
        assert sqlite_commit["is_segment"] is True

        # Ensure operation is persisted
        assert User.objects.using("postgres").exists()

        assert postgres_rollback["attributes"]["sentry.origin"] == "auto.http.django"
        assert postgres_commit["attributes"]["sentry.origin"] == "auto.http.django"
        assert sqlite_rollback["attributes"]["sentry.origin"] == "manual"
        assert sqlite_commit["attributes"]["sentry.origin"] == "manual"

        commit_spans = [
            span
            for span in spans
            if span["attributes"].get(SPANDATA.DB_OPERATION_NAME) == SPANNAME.DB_COMMIT
            or span["attributes"].get(SPANDATA.DB_OPERATION_NAME)
            == SPANNAME.DB_ROLLBACK
        ]
        assert len(commit_spans) == 0
    else:
        events = capture_events()

        client.get(reverse("postgres_insert_orm_no_autocommit_rollback"))
        client.get(reverse("postgres_insert_orm_no_autocommit"))

        with start_transaction(name="test_transaction"):
            from django.db import connection, transaction

            cursor = connection.cursor()

            query = """INSERT INTO auth_user (
    password,
    is_superuser,
    username,
    first_name,
    last_name,
    email,
    is_staff,
    is_active,
    date_joined
)
VALUES ('password', false, %s, %s, %s, %s, false, true, %s);"""

            query_list = (
                (
                    "user1",
                    "John",
                    "Doe",
                    "user1@example.com",
                    datetime(1970, 1, 1),
                ),
                (
                    "user2",
                    "Max",
                    "Mustermann",
                    "user2@example.com",
                    datetime(1970, 1, 1),
                ),
            )

            transaction.set_autocommit(False)
            cursor.executemany(query, query_list)
            transaction.rollback()
            transaction.set_autocommit(True)

        with start_transaction(name="test_transaction"):
            from django.db import connection, transaction

            cursor = connection.cursor()

            query = """INSERT INTO auth_user (
    password,
    is_superuser,
    username,
    first_name,
    last_name,
    email,
    is_staff,
    is_active,
    date_joined
)
VALUES ('password', false, %s, %s, %s, %s, false, true, %s);"""

            query_list = (
                (
                    "user1",
                    "John",
                    "Doe",
                    "user1@example.com",
                    datetime(1970, 1, 1),
                ),
                (
                    "user2",
                    "Max",
                    "Mustermann",
                    "user2@example.com",
                    datetime(1970, 1, 1),
                ),
            )

            transaction.set_autocommit(False)
            cursor.executemany(query, query_list)
            transaction.commit()
            transaction.set_autocommit(True)

        (postgres_rollback, postgres_commit, sqlite_rollback, sqlite_commit) = events

        # Ensure operation is persisted
        assert User.objects.using("postgres").exists()

        assert postgres_rollback["contexts"]["trace"]["origin"] == "auto.http.django"
        assert postgres_commit["contexts"]["trace"]["origin"] == "auto.http.django"
        assert sqlite_rollback["contexts"]["trace"]["origin"] == "manual"
        assert sqlite_commit["contexts"]["trace"]["origin"] == "manual"

        commit_spans = [
            span
            for span in itertools.chain(
                postgres_rollback["spans"],
                postgres_commit["spans"],
                sqlite_rollback["spans"],
                sqlite_commit["spans"],
            )
            if span["data"].get(SPANDATA.DB_OPERATION) == SPANNAME.DB_COMMIT
            or span["data"].get(SPANDATA.DB_OPERATION) == SPANNAME.DB_ROLLBACK
        ]
        assert len(commit_spans) == 0


@pytest.mark.forked
@pytest_mark_django_db_decorator(transaction=True)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_db_transaction_spans_disabled_atomic(
    sentry_init,
    client,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[DjangoIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if "postgres" not in connections:
        pytest.skip("postgres tests disabled")

    # trigger Django to open a new connection by marking the existing one as None.
    connections["postgres"].connection = None

    if span_streaming:
        items = capture_items("span")

        client.get(reverse("postgres_insert_orm_atomic_rollback"))
        client.get(reverse("postgres_insert_orm_atomic"))

        with sentry_sdk.traces.start_span(name="custom parent"):
            from django.db import connection, transaction

            with transaction.atomic():
                cursor = connection.cursor()

                query = """INSERT INTO auth_user (
    password,
    is_superuser,
    username,
    first_name,
    last_name,
    email,
    is_staff,
    is_active,
    date_joined
)
VALUES ('password', false, %s, %s, %s, %s, false, true, %s);"""

                query_list = (
                    (
                        "user1",
                        "John",
                        "Doe",
                        "user1@example.com",
                        datetime(1970, 1, 1),
                    ),
                    (
                        "user2",
                        "Max",
                        "Mustermann",
                        "user2@example.com",
                        datetime(1970, 1, 1),
                    ),
                )
                cursor.executemany(query, query_list)
                transaction.set_rollback(True)

        with sentry_sdk.traces.start_span(name="custom parent"):
            from django.db import connection, transaction

            with transaction.atomic():
                cursor = connection.cursor()

                query = """INSERT INTO auth_user (
    password,
    is_superuser,
    username,
    first_name,
    last_name,
    email,
    is_staff,
    is_active,
    date_joined
)
VALUES ('password', false, %s, %s, %s, %s, false, true, %s);"""

                query_list = (
                    (
                        "user1",
                        "John",
                        "Doe",
                        "user1@example.com",
                        datetime(1970, 1, 1),
                    ),
                    (
                        "user2",
                        "Max",
                        "Mustermann",
                        "user2@example.com",
                        datetime(1970, 1, 1),
                    ),
                )
                cursor.executemany(query, query_list)

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        postgres_rollback = spans[4]
        assert postgres_rollback["is_segment"] is True
        postgres_commit = spans[9]
        assert postgres_commit["is_segment"] is True
        sqlite_rollback = spans[12]
        assert sqlite_rollback["is_segment"] is True
        sqlite_commit = spans[15]
        assert sqlite_commit["is_segment"] is True

        commit_spans = [
            span
            for span in spans
            if span["attributes"].get(SPANDATA.DB_OPERATION_NAME) == SPANNAME.DB_COMMIT
            or span["attributes"].get(SPANDATA.DB_OPERATION_NAME)
            == SPANNAME.DB_ROLLBACK
        ]
    else:
        events = capture_events()

        client.get(reverse("postgres_insert_orm_atomic_rollback"))
        client.get(reverse("postgres_insert_orm_atomic"))

        with start_transaction(name="test_transaction"):
            from django.db import connection, transaction

            with transaction.atomic():
                cursor = connection.cursor()

                query = """INSERT INTO auth_user (
    password,
    is_superuser,
    username,
    first_name,
    last_name,
    email,
    is_staff,
    is_active,
    date_joined
)
VALUES ('password', false, %s, %s, %s, %s, false, true, %s);"""

                query_list = (
                    (
                        "user1",
                        "John",
                        "Doe",
                        "user1@example.com",
                        datetime(1970, 1, 1),
                    ),
                    (
                        "user2",
                        "Max",
                        "Mustermann",
                        "user2@example.com",
                        datetime(1970, 1, 1),
                    ),
                )
                cursor.executemany(query, query_list)
                transaction.set_rollback(True)

        with start_transaction(name="test_transaction"):
            from django.db import connection, transaction

            with transaction.atomic():
                cursor = connection.cursor()

                query = """INSERT INTO auth_user (
    password,
    is_superuser,
    username,
    first_name,
    last_name,
    email,
    is_staff,
    is_active,
    date_joined
)
VALUES ('password', false, %s, %s, %s, %s, false, true, %s);"""

                query_list = (
                    (
                        "user1",
                        "John",
                        "Doe",
                        "user1@example.com",
                        datetime(1970, 1, 1),
                    ),
                    (
                        "user2",
                        "Max",
                        "Mustermann",
                        "user2@example.com",
                        datetime(1970, 1, 1),
                    ),
                )
                cursor.executemany(query, query_list)

        (postgres_rollback, postgres_commit, sqlite_rollback, sqlite_commit) = events

        # Ensure operation is persisted
        assert User.objects.using("postgres").exists()

        assert postgres_rollback["contexts"]["trace"]["origin"] == "auto.http.django"
        assert postgres_commit["contexts"]["trace"]["origin"] == "auto.http.django"
        assert sqlite_rollback["contexts"]["trace"]["origin"] == "manual"
        assert sqlite_commit["contexts"]["trace"]["origin"] == "manual"

        commit_spans = [
            span
            for span in itertools.chain(
                postgres_rollback["spans"],
                postgres_commit["spans"],
                sqlite_rollback["spans"],
                sqlite_commit["spans"],
            )
            if span["data"].get(SPANDATA.DB_OPERATION) == SPANNAME.DB_COMMIT
            or span["data"].get(SPANDATA.DB_OPERATION) == SPANNAME.DB_ROLLBACK
        ]

    assert len(commit_spans) == 0


@pytest.mark.forked
@pytest_mark_django_db_decorator(transaction=True)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_db_no_autocommit_execute(
    sentry_init,
    client,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[DjangoIntegration(db_transaction_spans=True)],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if "postgres" not in connections:
        pytest.skip("postgres tests disabled")

    # trigger Django to open a new connection by marking the existing one as None.
    connections["postgres"].connection = None

    if span_streaming:
        items = capture_items("span")

        client.get(reverse("postgres_insert_orm_no_autocommit"))

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        # Ensure operation is persisted
        assert User.objects.using("postgres").exists()

        assert spans[5]["attributes"]["sentry.origin"] == "auto.http.django"

        commit_spans = [
            span
            for span in spans
            if span["attributes"].get(SPANDATA.DB_OPERATION_NAME) == SPANNAME.DB_COMMIT
        ]

        assert len(commit_spans) == 1
        commit_span = commit_spans[0]

        assert commit_span["attributes"]["sentry.origin"] == "auto.db.django"

        # Verify other database attributes
        assert commit_span["attributes"].get(SPANDATA.DB_SYSTEM_NAME) == "postgresql"
        conn_params = connections["postgres"].get_connection_params()
        assert commit_span["attributes"].get(SPANDATA.DB_NAMESPACE) is not None
        assert commit_span["attributes"].get(SPANDATA.DB_NAMESPACE) == conn_params.get(
            "database"
        ) or conn_params.get("dbname")
        assert commit_span["attributes"].get(SPANDATA.SERVER_ADDRESS) == os.environ.get(
            "SENTRY_PYTHON_TEST_POSTGRES_HOST", "localhost"
        )
        assert commit_span["attributes"].get(SPANDATA.SERVER_PORT) == os.environ.get(
            "SENTRY_PYTHON_TEST_POSTGRES_PORT", "5432"
        )

        insert_spans = [
            span for span in spans if span["name"].startswith("INSERT INTO")
        ]
    else:
        events = capture_events()

        client.get(reverse("postgres_insert_orm_no_autocommit"))

        (event,) = events

        # Ensure operation is persisted
        assert User.objects.using("postgres").exists()

        assert event["contexts"]["trace"]["origin"] == "auto.http.django"

        commit_spans = [
            span
            for span in event["spans"]
            if span["data"].get(SPANDATA.DB_OPERATION) == SPANNAME.DB_COMMIT
        ]

        assert len(commit_spans) == 1
        commit_span = commit_spans[0]

        assert commit_span["origin"] == "auto.db.django"

        # Verify other database attributes
        assert commit_span["data"].get(SPANDATA.DB_SYSTEM) == "postgresql"
        conn_params = connections["postgres"].get_connection_params()
        assert commit_span["data"].get(SPANDATA.DB_NAME) is not None
        assert commit_span["data"].get(SPANDATA.DB_NAME) == conn_params.get(
            "database"
        ) or conn_params.get("dbname")
        assert commit_span["data"].get(SPANDATA.SERVER_ADDRESS) == os.environ.get(
            "SENTRY_PYTHON_TEST_POSTGRES_HOST", "localhost"
        )
        assert commit_span["data"].get(SPANDATA.SERVER_PORT) == os.environ.get(
            "SENTRY_PYTHON_TEST_POSTGRES_PORT", "5432"
        )

        insert_spans = [
            span
            for span in event["spans"]
            if span["description"].startswith("INSERT INTO")
        ]
    assert len(insert_spans) == 1
    insert_span = insert_spans[0]

    # Verify query and commit statements are siblings
    assert commit_span["parent_span_id"] == insert_span["parent_span_id"]


@pytest.mark.forked
@pytest_mark_django_db_decorator(transaction=True)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_db_no_autocommit_executemany(
    sentry_init,
    client,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[DjangoIntegration(db_transaction_spans=True)],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            from django.db import connection, transaction

            cursor = connection.cursor()

            query = """INSERT INTO auth_user (
    password,
    is_superuser,
    username,
    first_name,
    last_name,
    email,
    is_staff,
    is_active,
    date_joined
)
VALUES ('password', false, %s, %s, %s, %s, false, true, %s);"""

            query_list = (
                (
                    "user1",
                    "John",
                    "Doe",
                    "user1@example.com",
                    datetime(1970, 1, 1),
                ),
                (
                    "user2",
                    "Max",
                    "Mustermann",
                    "user2@example.com",
                    datetime(1970, 1, 1),
                ),
            )

            transaction.set_autocommit(False)
            cursor.executemany(query, query_list)
            transaction.commit()
            transaction.set_autocommit(True)

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        # Ensure operation is persisted
        assert User.objects.exists()

        assert spans[2]["attributes"]["sentry.origin"] == "manual"
        assert spans[0]["attributes"]["sentry.origin"] == "auto.db.django"

        commit_spans = [
            span
            for span in spans
            if span["attributes"].get(SPANDATA.DB_OPERATION_NAME) == SPANNAME.DB_COMMIT
        ]

        assert len(commit_spans) == 1
        commit_span = commit_spans[0]

        assert commit_span["attributes"]["sentry.origin"] == "auto.db.django"

        # Verify other database attributes
        assert commit_span["attributes"].get(SPANDATA.DB_SYSTEM_NAME) == "sqlite"
        conn_params = connection.get_connection_params()
        assert commit_span["attributes"].get(SPANDATA.DB_NAMESPACE) is not None
        assert commit_span["attributes"].get(SPANDATA.DB_NAMESPACE) == conn_params.get(
            "database"
        ) or conn_params.get("dbname")

        insert_spans = [
            span for span in spans if span["name"].startswith("INSERT INTO")
        ]
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):
            from django.db import connection, transaction

            cursor = connection.cursor()

            query = """INSERT INTO auth_user (
    password,
    is_superuser,
    username,
    first_name,
    last_name,
    email,
    is_staff,
    is_active,
    date_joined
)
VALUES ('password', false, %s, %s, %s, %s, false, true, %s);"""

            query_list = (
                (
                    "user1",
                    "John",
                    "Doe",
                    "user1@example.com",
                    datetime(1970, 1, 1),
                ),
                (
                    "user2",
                    "Max",
                    "Mustermann",
                    "user2@example.com",
                    datetime(1970, 1, 1),
                ),
            )

            transaction.set_autocommit(False)
            cursor.executemany(query, query_list)
            transaction.commit()
            transaction.set_autocommit(True)

        (event,) = events

        # Ensure operation is persisted
        assert User.objects.exists()

        assert event["contexts"]["trace"]["origin"] == "manual"
        assert event["spans"][0]["origin"] == "auto.db.django"

        commit_spans = [
            span
            for span in event["spans"]
            if span["data"].get(SPANDATA.DB_OPERATION) == SPANNAME.DB_COMMIT
        ]

        assert len(commit_spans) == 1
        commit_span = commit_spans[0]

        assert commit_span["origin"] == "auto.db.django"

        # Verify other database attributes
        assert commit_span["data"].get(SPANDATA.DB_SYSTEM) == "sqlite"
        conn_params = connection.get_connection_params()
        assert commit_span["data"].get(SPANDATA.DB_NAME) is not None
        assert commit_span["data"].get(SPANDATA.DB_NAME) == conn_params.get(
            "database"
        ) or conn_params.get("dbname")

        insert_spans = [
            span
            for span in event["spans"]
            if span["description"].startswith("INSERT INTO")
        ]

    # Verify queries and commit statements are siblings
    for insert_span in insert_spans:
        assert commit_span["parent_span_id"] == insert_span["parent_span_id"]


@pytest.mark.forked
@pytest_mark_django_db_decorator(transaction=True)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_db_no_autocommit_rollback_execute(
    sentry_init,
    client,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[DjangoIntegration(db_transaction_spans=True)],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if "postgres" not in connections:
        pytest.skip("postgres tests disabled")

    # trigger Django to open a new connection by marking the existing one as None.
    connections["postgres"].connection = None

    if span_streaming:
        items = capture_items("span")

        client.get(reverse("postgres_insert_orm_no_autocommit_rollback"))

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        # Ensure operation is rolled back
        assert not User.objects.using("postgres").exists()

        assert spans[5]["attributes"]["sentry.origin"] == "auto.http.django"

        rollback_spans = [
            span
            for span in spans
            if span["attributes"].get(SPANDATA.DB_OPERATION_NAME)
            == SPANNAME.DB_ROLLBACK
        ]

        assert len(rollback_spans) == 1
        rollback_span = rollback_spans[0]

        assert rollback_span["attributes"]["sentry.origin"] == "auto.db.django"

        # Verify other database attributes
        assert rollback_span["attributes"].get(SPANDATA.DB_SYSTEM_NAME) == "postgresql"
        conn_params = connections["postgres"].get_connection_params()
        assert rollback_span["attributes"].get(SPANDATA.DB_NAMESPACE) is not None
        assert rollback_span["attributes"].get(
            SPANDATA.DB_NAMESPACE
        ) == conn_params.get("database") or conn_params.get("dbname")
        assert rollback_span["attributes"].get(
            SPANDATA.SERVER_ADDRESS
        ) == os.environ.get("SENTRY_PYTHON_TEST_POSTGRES_HOST", "localhost")
        assert rollback_span["attributes"].get(SPANDATA.SERVER_PORT) == os.environ.get(
            "SENTRY_PYTHON_TEST_POSTGRES_PORT", "5432"
        )

        insert_spans = [
            span for span in spans if span["name"].startswith("INSERT INTO")
        ]
    else:
        events = capture_events()

        client.get(reverse("postgres_insert_orm_no_autocommit_rollback"))

        (event,) = events

        # Ensure operation is rolled back
        assert not User.objects.using("postgres").exists()

        assert event["contexts"]["trace"]["origin"] == "auto.http.django"

        rollback_spans = [
            span
            for span in event["spans"]
            if span["data"].get(SPANDATA.DB_OPERATION) == SPANNAME.DB_ROLLBACK
        ]

        assert len(rollback_spans) == 1
        rollback_span = rollback_spans[0]

        assert rollback_span["origin"] == "auto.db.django"

        # Verify other database attributes
        assert rollback_span["data"].get(SPANDATA.DB_SYSTEM) == "postgresql"
        conn_params = connections["postgres"].get_connection_params()
        assert rollback_span["data"].get(SPANDATA.DB_NAME) is not None
        assert rollback_span["data"].get(SPANDATA.DB_NAME) == conn_params.get(
            "database"
        ) or conn_params.get("dbname")
        assert rollback_span["data"].get(SPANDATA.SERVER_ADDRESS) == os.environ.get(
            "SENTRY_PYTHON_TEST_POSTGRES_HOST", "localhost"
        )
        assert rollback_span["data"].get(SPANDATA.SERVER_PORT) == os.environ.get(
            "SENTRY_PYTHON_TEST_POSTGRES_PORT", "5432"
        )

        insert_spans = [
            span
            for span in event["spans"]
            if span["description"].startswith("INSERT INTO")
        ]
    assert len(insert_spans) == 1
    insert_span = insert_spans[0]

    # Verify query and rollback statements are siblings
    assert rollback_span["parent_span_id"] == insert_span["parent_span_id"]


@pytest.mark.forked
@pytest_mark_django_db_decorator(transaction=True)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_db_no_autocommit_rollback_executemany(
    sentry_init,
    client,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[DjangoIntegration(db_transaction_spans=True)],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            from django.db import connection, transaction

            cursor = connection.cursor()

            query = """INSERT INTO auth_user (
    password,
    is_superuser,
    username,
    first_name,
    last_name,
    email,
    is_staff,
    is_active,
    date_joined
)
VALUES ('password', false, %s, %s, %s, %s, false, true, %s);"""

            query_list = (
                (
                    "user1",
                    "John",
                    "Doe",
                    "user1@example.com",
                    datetime(1970, 1, 1),
                ),
                (
                    "user2",
                    "Max",
                    "Mustermann",
                    "user2@example.com",
                    datetime(1970, 1, 1),
                ),
            )

            transaction.set_autocommit(False)
            cursor.executemany(query, query_list)
            transaction.rollback()
            transaction.set_autocommit(True)

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        # Ensure operation is rolled back
        assert not User.objects.exists()

        assert spans[2]["attributes"]["sentry.origin"] == "manual"
        assert spans[0]["attributes"]["sentry.origin"] == "auto.db.django"

        rollback_spans = [
            span
            for span in spans
            if span["attributes"].get(SPANDATA.DB_OPERATION_NAME)
            == SPANNAME.DB_ROLLBACK
        ]
        assert len(rollback_spans) == 1
        rollback_span = rollback_spans[0]

        assert rollback_span["attributes"]["sentry.origin"] == "auto.db.django"

        # Verify other database attributes
        assert rollback_span["attributes"].get(SPANDATA.DB_SYSTEM_NAME) == "sqlite"
        conn_params = connection.get_connection_params()
        assert rollback_span["attributes"].get(SPANDATA.DB_NAMESPACE) is not None
        assert rollback_span["attributes"].get(
            SPANDATA.DB_NAMESPACE
        ) == conn_params.get("database") or conn_params.get("dbname")

        insert_spans = [
            span for span in spans if span["name"].startswith("INSERT INTO")
        ]
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):
            from django.db import connection, transaction

            cursor = connection.cursor()

            query = """INSERT INTO auth_user (
    password,
    is_superuser,
    username,
    first_name,
    last_name,
    email,
    is_staff,
    is_active,
    date_joined
)
VALUES ('password', false, %s, %s, %s, %s, false, true, %s);"""

            query_list = (
                (
                    "user1",
                    "John",
                    "Doe",
                    "user1@example.com",
                    datetime(1970, 1, 1),
                ),
                (
                    "user2",
                    "Max",
                    "Mustermann",
                    "user2@example.com",
                    datetime(1970, 1, 1),
                ),
            )

            transaction.set_autocommit(False)
            cursor.executemany(query, query_list)
            transaction.rollback()
            transaction.set_autocommit(True)

        (event,) = events

        # Ensure operation is rolled back
        assert not User.objects.exists()

        assert event["contexts"]["trace"]["origin"] == "manual"
        assert event["spans"][0]["origin"] == "auto.db.django"

        rollback_spans = [
            span
            for span in event["spans"]
            if span["data"].get(SPANDATA.DB_OPERATION) == SPANNAME.DB_ROLLBACK
        ]
        assert len(rollback_spans) == 1
        rollback_span = rollback_spans[0]

        assert rollback_span["origin"] == "auto.db.django"

        # Verify other database attributes
        assert rollback_span["data"].get(SPANDATA.DB_SYSTEM) == "sqlite"
        conn_params = connection.get_connection_params()
        assert rollback_span["data"].get(SPANDATA.DB_NAME) is not None
        assert rollback_span["data"].get(SPANDATA.DB_NAME) == conn_params.get(
            "database"
        ) or conn_params.get("dbname")

        insert_spans = [
            span
            for span in event["spans"]
            if span["description"].startswith("INSERT INTO")
        ]

    # Verify queries and rollback statements are siblings
    for insert_span in insert_spans:
        assert rollback_span["parent_span_id"] == insert_span["parent_span_id"]


@pytest.mark.forked
@pytest_mark_django_db_decorator(transaction=True)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_db_atomic_execute(
    sentry_init,
    client,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[DjangoIntegration(db_transaction_spans=True)],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if "postgres" not in connections:
        pytest.skip("postgres tests disabled")

    # trigger Django to open a new connection by marking the existing one as None.
    connections["postgres"].connection = None

    if span_streaming:
        items = capture_items("span")

        client.get(reverse("postgres_insert_orm_atomic"))

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        # Ensure operation is persisted
        assert User.objects.using("postgres").exists()

        assert spans[5]["attributes"]["sentry.origin"] == "auto.http.django"

        commit_spans = [
            span
            for span in spans
            if span["attributes"].get(SPANDATA.DB_OPERATION_NAME) == SPANNAME.DB_COMMIT
        ]

        assert len(commit_spans) == 1
        commit_span = commit_spans[0]

        assert commit_span["attributes"]["sentry.origin"] == "auto.db.django"

        # Verify other database attributes
        assert commit_span["attributes"].get(SPANDATA.DB_SYSTEM_NAME) == "postgresql"
        conn_params = connections["postgres"].get_connection_params()
        assert commit_span["attributes"].get(SPANDATA.DB_NAMESPACE) is not None
        assert commit_span["attributes"].get(SPANDATA.DB_NAMESPACE) == conn_params.get(
            "database"
        ) or conn_params.get("dbname")
        assert commit_span["attributes"].get(SPANDATA.SERVER_ADDRESS) == os.environ.get(
            "SENTRY_PYTHON_TEST_POSTGRES_HOST", "localhost"
        )
        assert commit_span["attributes"].get(SPANDATA.SERVER_PORT) == os.environ.get(
            "SENTRY_PYTHON_TEST_POSTGRES_PORT", "5432"
        )

        insert_spans = [
            span for span in spans if span["name"].startswith("INSERT INTO")
        ]
    else:
        events = capture_events()

        client.get(reverse("postgres_insert_orm_atomic"))

        (event,) = events

        # Ensure operation is persisted
        assert User.objects.using("postgres").exists()

        assert event["contexts"]["trace"]["origin"] == "auto.http.django"

        commit_spans = [
            span
            for span in event["spans"]
            if span["data"].get(SPANDATA.DB_OPERATION) == SPANNAME.DB_COMMIT
        ]

        assert len(commit_spans) == 1
        commit_span = commit_spans[0]

        assert commit_span["origin"] == "auto.db.django"

        # Verify other database attributes
        assert commit_span["data"].get(SPANDATA.DB_SYSTEM) == "postgresql"
        conn_params = connections["postgres"].get_connection_params()
        assert commit_span["data"].get(SPANDATA.DB_NAME) is not None
        assert commit_span["data"].get(SPANDATA.DB_NAME) == conn_params.get(
            "database"
        ) or conn_params.get("dbname")
        assert commit_span["data"].get(SPANDATA.SERVER_ADDRESS) == os.environ.get(
            "SENTRY_PYTHON_TEST_POSTGRES_HOST", "localhost"
        )
        assert commit_span["data"].get(SPANDATA.SERVER_PORT) == os.environ.get(
            "SENTRY_PYTHON_TEST_POSTGRES_PORT", "5432"
        )

        insert_spans = [
            span
            for span in event["spans"]
            if span["description"].startswith("INSERT INTO")
        ]
    assert len(insert_spans) == 1
    insert_span = insert_spans[0]

    # Verify query and commit statements are siblings
    assert commit_span["parent_span_id"] == insert_span["parent_span_id"]


@pytest.mark.forked
@pytest_mark_django_db_decorator(transaction=True)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_db_atomic_executemany(
    sentry_init,
    client,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[DjangoIntegration(db_transaction_spans=True)],
        send_default_pii=True,
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            from django.db import connection, transaction

            with transaction.atomic():
                cursor = connection.cursor()

                query = """INSERT INTO auth_user (
    password,
    is_superuser,
    username,
    first_name,
    last_name,
    email,
    is_staff,
    is_active,
    date_joined
)
VALUES ('password', false, %s, %s, %s, %s, false, true, %s);"""

                query_list = (
                    (
                        "user1",
                        "John",
                        "Doe",
                        "user1@example.com",
                        datetime(1970, 1, 1),
                    ),
                    (
                        "user2",
                        "Max",
                        "Mustermann",
                        "user2@example.com",
                        datetime(1970, 1, 1),
                    ),
                )
                cursor.executemany(query, query_list)

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        # Ensure operation is persisted
        assert User.objects.exists()

        assert spans[3]["attributes"]["sentry.origin"] == "manual"

        commit_spans = [
            span
            for span in spans
            if span["attributes"].get(SPANDATA.DB_OPERATION_NAME) == SPANNAME.DB_COMMIT
        ]
        assert len(commit_spans) == 1
        commit_span = commit_spans[0]

        assert commit_span["attributes"]["sentry.origin"] == "auto.db.django"

        # Verify other database attributes
        assert commit_span["attributes"].get(SPANDATA.DB_SYSTEM_NAME) == "sqlite"
        conn_params = connection.get_connection_params()
        assert commit_span["attributes"].get(SPANDATA.DB_NAMESPACE) is not None
        assert commit_span["attributes"].get(SPANDATA.DB_NAMESPACE) == conn_params.get(
            "database"
        ) or conn_params.get("dbname")

        insert_spans = [
            span for span in spans if span["name"].startswith("INSERT INTO")
        ]
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):
            from django.db import connection, transaction

            with transaction.atomic():
                cursor = connection.cursor()

                query = """INSERT INTO auth_user (
    password,
    is_superuser,
    username,
    first_name,
    last_name,
    email,
    is_staff,
    is_active,
    date_joined
)
VALUES ('password', false, %s, %s, %s, %s, false, true, %s);"""

                query_list = (
                    (
                        "user1",
                        "John",
                        "Doe",
                        "user1@example.com",
                        datetime(1970, 1, 1),
                    ),
                    (
                        "user2",
                        "Max",
                        "Mustermann",
                        "user2@example.com",
                        datetime(1970, 1, 1),
                    ),
                )
                cursor.executemany(query, query_list)

        (event,) = events

        # Ensure operation is persisted
        assert User.objects.exists()

        assert event["contexts"]["trace"]["origin"] == "manual"

        commit_spans = [
            span
            for span in event["spans"]
            if span["data"].get(SPANDATA.DB_OPERATION) == SPANNAME.DB_COMMIT
        ]
        assert len(commit_spans) == 1
        commit_span = commit_spans[0]

        assert commit_span["origin"] == "auto.db.django"

        # Verify other database attributes
        assert commit_span["data"].get(SPANDATA.DB_SYSTEM) == "sqlite"
        conn_params = connection.get_connection_params()
        assert commit_span["data"].get(SPANDATA.DB_NAME) is not None
        assert commit_span["data"].get(SPANDATA.DB_NAME) == conn_params.get(
            "database"
        ) or conn_params.get("dbname")

        insert_spans = [
            span
            for span in event["spans"]
            if span["description"].startswith("INSERT INTO")
        ]

    # Verify queries and commit statements are siblings
    for insert_span in insert_spans:
        assert commit_span["parent_span_id"] == insert_span["parent_span_id"]


@pytest.mark.forked
@pytest_mark_django_db_decorator(transaction=True)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_db_atomic_rollback_execute(
    sentry_init,
    client,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[DjangoIntegration(db_transaction_spans=True)],
        send_default_pii=True,
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if "postgres" not in connections:
        pytest.skip("postgres tests disabled")

    # trigger Django to open a new connection by marking the existing one as None.
    connections["postgres"].connection = None

    if span_streaming:
        items = capture_items("span")

        client.get(reverse("postgres_insert_orm_atomic_rollback"))

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        # Ensure operation is rolled back
        assert not User.objects.using("postgres").exists()

        assert spans[5]["attributes"]["sentry.origin"] == "auto.http.django"

        rollback_spans = [
            span
            for span in spans
            if span["attributes"].get(SPANDATA.DB_OPERATION_NAME)
            == SPANNAME.DB_ROLLBACK
        ]

        assert len(rollback_spans) == 1
        rollback_span = rollback_spans[0]

        assert rollback_span["attributes"]["sentry.origin"] == "auto.db.django"

        # Verify other database attributes
        assert rollback_span["attributes"].get(SPANDATA.DB_SYSTEM_NAME) == "postgresql"
        conn_params = connections["postgres"].get_connection_params()
        assert rollback_span["attributes"].get(SPANDATA.DB_NAMESPACE) is not None
        assert rollback_span["attributes"].get(
            SPANDATA.DB_NAMESPACE
        ) == conn_params.get("database") or conn_params.get("dbname")
        assert rollback_span["attributes"].get(
            SPANDATA.SERVER_ADDRESS
        ) == os.environ.get("SENTRY_PYTHON_TEST_POSTGRES_HOST", "localhost")
        assert rollback_span["attributes"].get(SPANDATA.SERVER_PORT) == os.environ.get(
            "SENTRY_PYTHON_TEST_POSTGRES_PORT", "5432"
        )

        insert_spans = [
            span for span in spans if span["name"].startswith("INSERT INTO")
        ]
    else:
        events = capture_events()

        client.get(reverse("postgres_insert_orm_atomic_rollback"))

        (event,) = events

        # Ensure operation is rolled back
        assert not User.objects.using("postgres").exists()

        assert event["contexts"]["trace"]["origin"] == "auto.http.django"

        rollback_spans = [
            span
            for span in event["spans"]
            if span["data"].get(SPANDATA.DB_OPERATION) == SPANNAME.DB_ROLLBACK
        ]

        assert len(rollback_spans) == 1
        rollback_span = rollback_spans[0]

        assert rollback_span["origin"] == "auto.db.django"

        # Verify other database attributes
        assert rollback_span["data"].get(SPANDATA.DB_SYSTEM) == "postgresql"
        conn_params = connections["postgres"].get_connection_params()
        assert rollback_span["data"].get(SPANDATA.DB_NAME) is not None
        assert rollback_span["data"].get(SPANDATA.DB_NAME) == conn_params.get(
            "database"
        ) or conn_params.get("dbname")
        assert rollback_span["data"].get(SPANDATA.SERVER_ADDRESS) == os.environ.get(
            "SENTRY_PYTHON_TEST_POSTGRES_HOST", "localhost"
        )
        assert rollback_span["data"].get(SPANDATA.SERVER_PORT) == os.environ.get(
            "SENTRY_PYTHON_TEST_POSTGRES_PORT", "5432"
        )

        insert_spans = [
            span
            for span in event["spans"]
            if span["description"].startswith("INSERT INTO")
        ]

    assert len(insert_spans) == 1
    insert_span = insert_spans[0]

    # Verify query and rollback statements are siblings
    assert rollback_span["parent_span_id"] == insert_span["parent_span_id"]


@pytest.mark.forked
@pytest_mark_django_db_decorator(transaction=True)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_db_atomic_rollback_executemany(
    sentry_init,
    client,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[DjangoIntegration(db_transaction_spans=True)],
        send_default_pii=True,
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            from django.db import connection, transaction

            with transaction.atomic():
                cursor = connection.cursor()

                query = """INSERT INTO auth_user (
    password,
    is_superuser,
    username,
    first_name,
    last_name,
    email,
    is_staff,
    is_active,
    date_joined
)
VALUES ('password', false, %s, %s, %s, %s, false, true, %s);"""

                query_list = (
                    (
                        "user1",
                        "John",
                        "Doe",
                        "user1@example.com",
                        datetime(1970, 1, 1),
                    ),
                    (
                        "user2",
                        "Max",
                        "Mustermann",
                        "user2@example.com",
                        datetime(1970, 1, 1),
                    ),
                )
                cursor.executemany(query, query_list)
                transaction.set_rollback(True)

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        # Ensure operation is rolled back
        assert not User.objects.exists()

        assert spans[3]["attributes"]["sentry.origin"] == "manual"

        rollback_spans = [
            span
            for span in spans
            if span["attributes"].get(SPANDATA.DB_OPERATION_NAME)
            == SPANNAME.DB_ROLLBACK
        ]

        assert len(rollback_spans) == 1
        rollback_span = rollback_spans[0]

        assert rollback_span["attributes"]["sentry.origin"] == "auto.db.django"

        # Verify other database attributes
        assert rollback_span["attributes"].get(SPANDATA.DB_SYSTEM_NAME) == "sqlite"
        conn_params = connection.get_connection_params()
        assert rollback_span["attributes"].get(SPANDATA.DB_NAMESPACE) is not None
        assert rollback_span["attributes"].get(
            SPANDATA.DB_NAMESPACE
        ) == conn_params.get("database") or conn_params.get("dbname")

        insert_spans = [
            span for span in spans if span["name"].startswith("INSERT INTO")
        ]
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):
            from django.db import connection, transaction

            with transaction.atomic():
                cursor = connection.cursor()

                query = """INSERT INTO auth_user (
    password,
    is_superuser,
    username,
    first_name,
    last_name,
    email,
    is_staff,
    is_active,
    date_joined
)
VALUES ('password', false, %s, %s, %s, %s, false, true, %s);"""

                query_list = (
                    (
                        "user1",
                        "John",
                        "Doe",
                        "user1@example.com",
                        datetime(1970, 1, 1),
                    ),
                    (
                        "user2",
                        "Max",
                        "Mustermann",
                        "user2@example.com",
                        datetime(1970, 1, 1),
                    ),
                )
                cursor.executemany(query, query_list)
                transaction.set_rollback(True)

        (event,) = events

        # Ensure operation is rolled back
        assert not User.objects.exists()

        assert event["contexts"]["trace"]["origin"] == "manual"

        rollback_spans = [
            span
            for span in event["spans"]
            if span["data"].get(SPANDATA.DB_OPERATION) == SPANNAME.DB_ROLLBACK
        ]

        assert len(rollback_spans) == 1
        rollback_span = rollback_spans[0]

        assert rollback_span["origin"] == "auto.db.django"

        # Verify other database attributes
        assert rollback_span["data"].get(SPANDATA.DB_SYSTEM) == "sqlite"
        conn_params = connection.get_connection_params()
        assert rollback_span["data"].get(SPANDATA.DB_NAME) is not None
        assert rollback_span["data"].get(SPANDATA.DB_NAME) == conn_params.get(
            "database"
        ) or conn_params.get("dbname")

        insert_spans = [
            span
            for span in event["spans"]
            if span["description"].startswith("INSERT INTO")
        ]

    # Verify queries and rollback statements are siblings
    for insert_span in insert_spans:
        assert rollback_span["parent_span_id"] == insert_span["parent_span_id"]


@pytest.mark.forked
@pytest_mark_django_db_decorator(transaction=True)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_db_atomic_execute_exception(
    sentry_init,
    client,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[DjangoIntegration(db_transaction_spans=True)],
        send_default_pii=True,
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if "postgres" not in connections:
        pytest.skip("postgres tests disabled")

    # trigger Django to open a new connection by marking the existing one as None.
    connections["postgres"].connection = None

    if span_streaming:
        items = capture_items("span")

        client.get(reverse("postgres_insert_orm_atomic_exception"))

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        # Ensure operation is rolled back
        assert not User.objects.using("postgres").exists()

        assert spans[5]["attributes"]["sentry.origin"] == "auto.http.django"

        rollback_spans = [
            span
            for span in spans
            if span["attributes"].get(SPANDATA.DB_OPERATION_NAME)
            == SPANNAME.DB_ROLLBACK
        ]

        assert len(rollback_spans) == 1
        rollback_span = rollback_spans[0]

        assert rollback_span["attributes"]["sentry.origin"] == "auto.db.django"

        # Verify other database attributes
        assert rollback_span["attributes"].get(SPANDATA.DB_SYSTEM_NAME) == "postgresql"
        conn_params = connections["postgres"].get_connection_params()
        assert rollback_span["attributes"].get(SPANDATA.DB_NAMESPACE) is not None
        assert rollback_span["attributes"].get(
            SPANDATA.DB_NAMESPACE
        ) == conn_params.get("database") or conn_params.get("dbname")
        assert rollback_span["attributes"].get(
            SPANDATA.SERVER_ADDRESS
        ) == os.environ.get("SENTRY_PYTHON_TEST_POSTGRES_HOST", "localhost")
        assert rollback_span["attributes"].get(SPANDATA.SERVER_PORT) == os.environ.get(
            "SENTRY_PYTHON_TEST_POSTGRES_PORT", "5432"
        )

        insert_spans = [
            span for span in spans if span["name"].startswith("INSERT INTO")
        ]
    else:
        events = capture_events()

        client.get(reverse("postgres_insert_orm_atomic_exception"))

        (event,) = events

        # Ensure operation is rolled back
        assert not User.objects.using("postgres").exists()

        assert event["contexts"]["trace"]["origin"] == "auto.http.django"

        rollback_spans = [
            span
            for span in event["spans"]
            if span["data"].get(SPANDATA.DB_OPERATION) == SPANNAME.DB_ROLLBACK
        ]

        assert len(rollback_spans) == 1
        rollback_span = rollback_spans[0]

        assert rollback_span["origin"] == "auto.db.django"

        # Verify other database attributes
        assert rollback_span["data"].get(SPANDATA.DB_SYSTEM) == "postgresql"
        conn_params = connections["postgres"].get_connection_params()
        assert rollback_span["data"].get(SPANDATA.DB_NAME) is not None
        assert rollback_span["data"].get(SPANDATA.DB_NAME) == conn_params.get(
            "database"
        ) or conn_params.get("dbname")
        assert rollback_span["data"].get(SPANDATA.SERVER_ADDRESS) == os.environ.get(
            "SENTRY_PYTHON_TEST_POSTGRES_HOST", "localhost"
        )
        assert rollback_span["data"].get(SPANDATA.SERVER_PORT) == os.environ.get(
            "SENTRY_PYTHON_TEST_POSTGRES_PORT", "5432"
        )

        insert_spans = [
            span
            for span in event["spans"]
            if span["description"].startswith("INSERT INTO")
        ]
    assert len(insert_spans) == 1
    insert_span = insert_spans[0]

    # Verify query and rollback statements are siblings
    assert rollback_span["parent_span_id"] == insert_span["parent_span_id"]


@pytest.mark.forked
@pytest_mark_django_db_decorator(transaction=True)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_db_atomic_executemany_exception(
    sentry_init,
    client,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[DjangoIntegration(db_transaction_spans=True)],
        send_default_pii=True,
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            from django.db import connection, transaction

            try:
                with transaction.atomic():
                    cursor = connection.cursor()

                    query = """INSERT INTO auth_user (
    password,
    is_superuser,
    username,
    first_name,
    last_name,
    email,
    is_staff,
    is_active,
    date_joined
)
VALUES ('password', false, %s, %s, %s, %s, false, true, %s);"""

                    query_list = (
                        (
                            "user1",
                            "John",
                            "Doe",
                            "user1@example.com",
                            datetime(1970, 1, 1),
                        ),
                        (
                            "user2",
                            "Max",
                            "Mustermann",
                            "user2@example.com",
                            datetime(1970, 1, 1),
                        ),
                    )
                    cursor.executemany(query, query_list)
                    1 / 0
            except ZeroDivisionError:
                pass

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        # Ensure operation is rolled back
        assert not User.objects.exists()

        assert spans[3]["attributes"]["sentry.origin"] == "manual"

        rollback_spans = [
            span
            for span in spans
            if span["attributes"].get(SPANDATA.DB_OPERATION_NAME)
            == SPANNAME.DB_ROLLBACK
        ]
        assert len(rollback_spans) == 1
        rollback_span = rollback_spans[0]

        assert rollback_span["attributes"]["sentry.origin"] == "auto.db.django"

        # Verify other database attributes
        assert rollback_span["attributes"].get(SPANDATA.DB_SYSTEM_NAME) == "sqlite"
        conn_params = connection.get_connection_params()
        assert rollback_span["attributes"].get(SPANDATA.DB_NAMESPACE) is not None
        assert rollback_span["attributes"].get(
            SPANDATA.DB_NAMESPACE
        ) == conn_params.get("database") or conn_params.get("dbname")

        insert_spans = [
            span for span in spans if span["name"].startswith("INSERT INTO")
        ]
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):
            from django.db import connection, transaction

            try:
                with transaction.atomic():
                    cursor = connection.cursor()

                    query = """INSERT INTO auth_user (
    password,
    is_superuser,
    username,
    first_name,
    last_name,
    email,
    is_staff,
    is_active,
    date_joined
)
VALUES ('password', false, %s, %s, %s, %s, false, true, %s);"""

                    query_list = (
                        (
                            "user1",
                            "John",
                            "Doe",
                            "user1@example.com",
                            datetime(1970, 1, 1),
                        ),
                        (
                            "user2",
                            "Max",
                            "Mustermann",
                            "user2@example.com",
                            datetime(1970, 1, 1),
                        ),
                    )
                    cursor.executemany(query, query_list)
                    1 / 0
            except ZeroDivisionError:
                pass

        (event,) = events

        # Ensure operation is rolled back
        assert not User.objects.exists()

        assert event["contexts"]["trace"]["origin"] == "manual"

        rollback_spans = [
            span
            for span in event["spans"]
            if span["data"].get(SPANDATA.DB_OPERATION) == SPANNAME.DB_ROLLBACK
        ]
        assert len(rollback_spans) == 1
        rollback_span = rollback_spans[0]

        assert rollback_span["origin"] == "auto.db.django"

        # Verify other database attributes
        assert rollback_span["data"].get(SPANDATA.DB_SYSTEM) == "sqlite"
        conn_params = connection.get_connection_params()
        assert rollback_span["data"].get(SPANDATA.DB_NAME) is not None
        assert rollback_span["data"].get(SPANDATA.DB_NAME) == conn_params.get(
            "database"
        ) or conn_params.get("dbname")

        insert_spans = [
            span
            for span in event["spans"]
            if span["description"].startswith("INSERT INTO")
        ]

    # Verify queries and rollback statements are siblings
    for insert_span in insert_spans:
        assert rollback_span["parent_span_id"] == insert_span["parent_span_id"]
